/*-------------------------------------------------------------------------
 * Copyright (c) 2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_analyze
 *
 * Synopsis:
 *    
 * $Id: db_analyze.c,v 1.5 2004/06/29 18:45:41 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <gnome.h>
#include <libpq-fe.h>

#include "db_login.h"
#include "db_schema.h"

#define TABLE_EXISTS "\
select relname \
  from pg_class \
 where relkind = 'r' \
   and relname = '%s'"

#define TABLE_PK_EXISTS "\
select c.conname \
  from pg_constraint c, pg_class t \
 where c.contype = 'p' \
   and c.conrelid = t.oid \
   and t.relname = '%s'"

#define COLUMN_EXISTS "\
select a.attname \
  from pg_attribute a, pg_class c \
 where c.oid = a.attrelid \
   and c.relname = '%s' \
   and a.attname = '%s'"

#define LANG_EXISTS "\
select lanname \
  from pg_language \
 where lanname = '%s'"

#define INDEX_EXISTS "\
select relname \
  from pg_class \
 where relkind = 'i' \
   and relname = '%s'"

#define GET_VERSION "select max (schema_ver) from version"

static GList *missing_tables = NULL;
static GList *missing_columns = NULL;

gboolean
table_exists (PGconn *conn, const gchar *table)
{
   GString  *sql;
   gboolean  exists = FALSE;
   PGresult *res;

   sql = g_string_new ("");
   g_string_printf (sql, TABLE_EXISTS, table);
   res = PQexec (conn, sql->str);
   if (chk_sql_ok (res, _("Determine if table exists")))
      {
      exists = (PQntuples (res) != 0);
      }
   PQclear (res);

   g_string_free (sql, TRUE);
   return exists;
}


gboolean
table_pk_exists (PGconn *conn, const gchar *table)
{
   GString    *sql;
   gboolean    exists = FALSE;
   PGresult   *res;

   sql = g_string_new ("");
   g_string_printf (sql, TABLE_PK_EXISTS, table);
   res = PQexec (conn, sql->str);
   if (chk_sql_ok (res, _("Determine if table exists")))
      {
      exists = (PQntuples (res) != 0);
      }
   PQclear (res);

   g_string_free (sql, TRUE);
   return exists;
}


gboolean
col_exists (PGconn *conn, const gchar *table, const gchar *col)
{
   GString  *sql;
   gboolean  exists = FALSE;
   PGresult *res;

   sql = g_string_new ("");
   g_string_printf (sql, COLUMN_EXISTS, table, col);
   res = PQexec (conn, sql->str);
   if (chk_sql_ok (res, _("Determine if column exists")))
      {
      exists = (PQntuples (res) != 0);
      }
   PQclear (res);

   g_string_free (sql, TRUE);
   return exists;
}


gboolean
lang_exists (PGconn *conn, const gchar *lang)
{
   GString  *sql;
   gboolean  exists = FALSE;
   PGresult *res;

   sql = g_string_new ("");
   g_string_printf (sql, LANG_EXISTS, lang);
   res = PQexec (conn, sql->str);
   if (chk_sql_ok (res, _("Determine if language exists")))
      {
      exists = (PQntuples (res) != 0);
      }
   PQclear (res);

   g_string_free (sql, TRUE);
   return exists;
}


gboolean
index_exists (PGconn *conn, const gchar *idx_name)
{
   GString   *sql;
   gboolean   exists = FALSE;
   PGresult  *res;

   sql = g_string_new ("");
   g_string_printf (sql, INDEX_EXISTS, idx_name);
   res = PQexec (conn, sql->str);
   if (chk_sql_ok (res, _("Determine if index exists")))
      {
      exists = (PQntuples (res) != 0);
      }
   PQclear (res);

   g_string_free (sql, TRUE);
   return exists;
}


gint
get_db_version (PGconn *conn)
{
   gint       ver = 0;
   PGresult  *res = NULL;

   if (col_exists (conn, "version", "schema_ver"))
      {
      res = PQexec (conn, GET_VERSION);
      if (chk_sql_ok (res, _("Determine the version of the schema")))
         {
         if (PQntuples (res) == 1)
            {
            ver = atoi (PQgetvalue (res, 0, 0));
            }
         }
      PQclear (res);
      }

   return ver;
}


gboolean
db_has_prereq_objects (PGconn *conn)
{
   TableData   *table;
   gint         i;
   gint         n;
   gboolean     valid = TRUE;

   /*
    * If there are lists hanging around from a previous analysis, clear
    * them.
    */
   if (missing_tables != NULL)
      {
      g_list_free (missing_tables);
      missing_tables = NULL;
      }

   if (missing_columns != NULL)
      {
      g_list_free (missing_columns);
      missing_columns = NULL;
      }

   /*
    * Tables might not be required, but may have columns that are required
    * if the table exists.  Check all defined tables.  If the table exists,
    * make sure all required columns exist.  If the table does not exist,
    * just make sure the table itself is not a required table.
    */
   n = count_tables();
   for (i = 0; i < n; i++)
      {
      table = get_table (i);
      if (table_exists (conn, table->name->str))
         {
         GList *iter;

         for (iter = g_list_first (table->columns);
               iter != NULL;
               iter = iter->next)
            {
            ColumnData *col = iter->data;

            if (col->prereq && !col_exists (conn, table->name->str,
                                            col->name->str))
               {
               missing_columns = g_list_append (missing_columns, table->name);
               missing_columns = g_list_append (missing_columns, col->name);
               valid = FALSE;
               }
            }
         } 
      else
         {
         if (table->prereq)
            {
            missing_tables = g_list_append (missing_tables, table->name);
            valid = FALSE;
            }
         }
      }

   return valid;
}
