/*-------------------------------------------------------------------------
 * Copyright (c) 2003-2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_create
 *
 * Synopsis:
 *    Functions used for database creation.
 *    
 * $Id: db_create.c,v 1.11 2004/04/18 21:42:21 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "db_create.h"
#include "db_ddl.h"
#include "db_schema.h"


PGresult *
create_new_database (DbLoginData *newdb_data, DbLoginData *pgsu_data)
{
   PGresult *res = NULL;
   GString  *sql;

   sql = g_string_new ("CREATE DATABASE ");
   g_string_append_printf (sql, "%s WITH OWNER = %s",
                           newdb_data->db_name->str,
                           newdb_data->username->str);
   res = PQexec (pgsu_data->conn, sql->str);
   g_string_free (sql, TRUE);
   return res;
}


PGresult *
create_sequences (DbLoginData *data, gint version)
{
   SequenceData   *seq;
   PGresult       *res = NULL;
   ExecStatusType  status;
   GString        *sql;
   gint            i;
   gint            n;

   /*
    * TODO: The version is not used at this point.  It would be good to
    *       add a minver to the sequence structure.
    */

   sql = g_string_new ("");
   n = count_sequences();
   for (i = 0; i < n; i++)
      {
      seq = get_sequence (i);
      g_string_printf (sql, CR8SEQ, seq->name->str, seq->start,
                       seq->increment, seq->maxval, seq->minval,
                       seq->cache);
      if (res != NULL)
         {
         PQclear (res);
         }
      res = PQexec (data->conn, sql->str);
      status = PQresultStatus (res);
      if (status != PGRES_COMMAND_OK)
         {
         break;
         }
      PQclear (res);
      res = fix_grants (data->conn, seq->name->str);
      if (status != PGRES_COMMAND_OK)
         {
         break;
         }
      }

   g_string_free (sql, TRUE);
   return res;
}


PGresult *
create_tables (DbLoginData *data, gint version)
{
   GString        *sql;
   TableData      *tab;
   ExecStatusType  status;
   PGresult       *res = NULL;
   gint            i;
   gint            n;

   sql = g_string_new ("");
   n = count_tables ();
   for (i = 0; i < n; i++)
      {
      tab = get_table (i);
      if (tab->minver <= version)
         {
         create_table_str (sql, i, version);
         if (res != NULL)
            {
            PQclear (res);
            }
         res = PQexec (data->conn, sql->str);
         status = PQresultStatus (res);
         if (status != PGRES_COMMAND_OK)
            {
            break;
            }
         PQclear (res);
         res = fix_grants (data->conn, tab->name->str);
         if (status != PGRES_COMMAND_OK)
            {
            break;
            }
         }
      }

   g_string_free (sql, TRUE);
   return res;
}


PGresult *
create_functions (DbLoginData *data, gint version)
{
   GString        *sql;
   PGresult       *res = NULL;
   ExecStatusType  status;
   FunctionData   *func;
   gint            i;
   gint            n;

   sql = g_string_new ("");
   n = count_functions ();
   for (i = 0; i < n; i++)
      {
      func = get_function (i);
      /*
       * Check if this function should be included in this version of
       * the PRepS schema.
       */
      if (version < func->minver ||
            (func->maxver != NOVER && version > func->maxver))
         {
         continue;
         }

      /*
       * Check to see if the funtion already exists (some of these
       * items could have been loaded into the template schema).
       */
      g_string_printf (sql,
                       "select proname from pg_proc where proname = '%s'",
                       func->name->str);
      if (res != NULL)
         {
         PQclear (res);
         }
      res = PQexec (data->conn, sql->str);
      status = PQresultStatus (res);
      if (status != PGRES_TUPLES_OK && status != PGRES_COMMAND_OK)
         {
         break;
         }
      /*
       * Only create the function is it does not already exist.
       */
      if (PQntuples (res) == 0)
         {
         g_string_printf
            (sql,
             "CREATE FUNCTION \"%s\" (%s) RETURNS %s AS '%s' LANGUAGE '%s'",
             func->name->str, func->args->str, func->rtn->str,
             func->libpath->str, func->lang->str);
         res = PQexec (data->conn, sql->str);
         status = PQresultStatus (res);
         if (status != PGRES_TUPLES_OK && status != PGRES_COMMAND_OK)
            {
            break;
            }
         }
      }

   g_string_free (sql, TRUE); 
   return res;
}


PGresult *
create_languages (DbLoginData *data, gint version)
{
   GString        *sql;
   PGresult       *res = NULL;
   ExecStatusType  status;
   LanguageData   *lang;
   gint            i;
   gint            n;

   sql = g_string_new ("");
   n = count_languages ();
   for (i = 0; i < n; i++)
      {
      lang = get_language (i);
      /*
       * Check if this language should be included in this version of
       * the PRepS schema.
       */
      if (version < lang->minver ||
            (lang->maxver != NOVER && version > lang->maxver))
         {
         continue;
         }

      /*
       * Check to see if the funtion already exists (some of these
       * items could have been loaded into the template schema).
       */
      g_string_printf (sql,
                       "select lanname from pg_language where lanname = '%s'",
                       lang->name->str);
      if (res != NULL)
         {
         PQclear (res);
         }
      res = PQexec (data->conn, sql->str);
      status = PQresultStatus (res);
      if (status != PGRES_TUPLES_OK && status != PGRES_COMMAND_OK)
         {
         break;
         }
      /*
       * Only create the language if it does not already exist.
       */
      if (PQntuples (res) == 0)
         {
         g_string_printf (sql,
                          "CREATE TRUSTED PROCEDURAL LANGUAGE '%s' HANDLER \"%s\" LANCOMPILER '%s'",
                          lang->name->str, lang->handler->str,
                          lang->compiler->str);
         PQclear (res);
         res = PQexec (data->conn, sql->str);
         status = PQresultStatus (res);
         if (status != PGRES_TUPLES_OK && status != PGRES_COMMAND_OK)
            {
            break;
            }
         }
      }

   g_string_free (sql, TRUE); 
   return res;
}


PGresult *
create_stored_funcs (DbLoginData *data, gint version)
{
   GString        *sql;
   PGresult       *res = NULL;
   ExecStatusType  status;
   StoredFuncData *sf;
   gint            i;
   gint            n;

   sql = g_string_new ("");
   n = count_stored_funcs ();
   for (i = 0; i < n; i++)
      {
      sf = get_stored_func (i);
      /*
       * Check if this function should be included in this version of
       * the PRepS schema.
       */
      if (version < sf->minver ||
            (sf->maxver != NOVER && version > sf->maxver))
         {
         continue;
         }

      g_string_printf
         (sql,
          "CREATE FUNCTION \"%s\" (%s) RETURNS %s AS '%s' LANGUAGE '%s'",
          sf->name->str, sf->args->str, sf->rtn->str, sf->code->str,
          sf->lang->str);
      if (res != NULL)
         {
         PQclear (res);
         }
      res = PQexec (data->conn, sql->str);
      status = PQresultStatus (res);
      if (status != PGRES_COMMAND_OK)
         {
         break;
         }
      }

   g_string_free (sql, TRUE);
   return res;
}


PGresult *
create_unique_indexes (DbLoginData *data, gint version)
{
   GString         *sql;
   PGresult        *res = NULL;
   ExecStatusType   status;
   UniqueIndexData *ui;
   gint             i;
   gint             n;

   sql = g_string_new ("");
   n = count_unique_indexes ();
   for (i = 0; i < n; i++)
      {
      ui = get_unique_index (i);
      /*
       * Check if this index should be included in this version of
       * the PRepS schema.
       */
      if (version < ui->minver ||
            (ui->maxver != NOVER && version > ui->maxver))
         {
         continue;
         }
      
      g_string_printf (sql,
                       "CREATE UNIQUE INDEX \"%s\" ON \"%s\" USING BTREE ( \"%s\" \"text_ops\" )",
                       ui->name->str, ui->table->str, ui->column->str);
      if (res != NULL)
         {
         PQclear (res);
         }
      res = PQexec (data->conn, sql->str);
      status = PQresultStatus (res);
      if (status != PGRES_COMMAND_OK)
         {
         break;
         }
      }

   g_string_free (sql, TRUE);
   return res;
}


PGresult *
create_triggers (DbLoginData *data, gint version)
{
   GString         *sql;
   PGresult        *res = NULL;
   ExecStatusType   status;
   TriggerData     *td;
   gint             i;
   gint             n;

   sql = g_string_new ("");
   n = count_triggers ();
   for (i = 0; i < n; i++)
      {
      td = get_trigger (i);
      /*
       * Check if this trigger should be included in this version of
       * the PRepS schema.
       */
      if (version < td->minver ||
            (td->maxver != NOVER && version > td->maxver))
         {
         continue;
         }
      
      g_string_printf (sql, "CREATE TRIGGER \"%s\" ", td->name->str);
      sql = g_string_append (sql, td->before ? "BEFORE " : "AFTER ");
      if (td->ins)
         {
         sql = g_string_append (sql, "INSERT OR ");
         }
      if (td->upd)
         {
         sql = g_string_append (sql, "UPDATE OR ");
         }
      if (td->del)
         {
         sql = g_string_append (sql, "DELETE OR ");
         }
      sql = g_string_truncate (sql, sql->len - 3);
      g_string_append_printf (sql, "ON \"%s\" FOR EACH ", td->table->str);
      sql = g_string_append (sql, td->row ? "ROW " : "STATEMENT ");
      g_string_append_printf (sql, "EXECUTE PROCEDURE \"%s\" (%s)",
                              td->func->str, td->args->str);
      if (res != NULL)
         {
         PQclear (res);
         }
      res = PQexec (data->conn, sql->str);
      status = PQresultStatus (res);
      if (status != PGRES_COMMAND_OK)
         {
         break;
         }
      }

   g_string_free (sql, TRUE);
   return res;
}


PGresult *
create_rules (DbLoginData *data, gint version)
{
   GString         *sql;
   PGresult        *res = NULL;
   ExecStatusType   status;
   RuleData        *rd;
   gint             i;
   gint             n;

   sql = g_string_new ("");
   n = count_rules ();
   for (i = 0; i < n; i++)
      {
      rd = get_rule (i);
      /*
       * Check if this rule should be included in this version of
       * the PRepS schema.
       */
      if (version < rd->minver ||
            (rd->maxver != NOVER && version > rd->maxver))
         {
         continue;
         }
      
      g_string_printf (sql, "CREATE OR REPLACE RULE \"%s\" AS ON ",
                       rd->name->str);
      switch (rd->event)
         {
         case SELECT_EVENT:
            sql = g_string_append (sql, "SELECT ");
            break;
         case UPDATE_EVENT:
            sql = g_string_append (sql, "UPDATE ");
            break;
         case INSERT_EVENT:
            sql = g_string_append (sql, "INSERT ");
            break;
         case DELETE_EVENT:
            sql = g_string_append (sql, "DELETE ");
            break;
         default:
            g_assert_not_reached();
         }
      g_string_append_printf (sql, "TO %s WHERE %s DO %s %s",
                              rd->table->str, rd->where->str,
                              rd->instead ? "INSTEAD" : "",
                              rd->action->str);
      if (res != NULL)
         {
         PQclear (res);
         }
      res = PQexec (data->conn, sql->str);
      status = PQresultStatus (res);
      if (status != PGRES_COMMAND_OK)
         {
         break;
         }
      }

   g_string_free (sql, TRUE);
   return res;
}
