#ifndef __DB_CREATE_H
#define __DB_CREATE_H
/*-------------------------------------------------------------------------
 * Copyright (c) 2003-2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_create
 *
 * Synopsis:
 *    Functions used for database creation.
 *    
 * $Id: db_create.h,v 1.10 2004/04/18 21:42:21 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#include <libpq-fe.h>
#include <glib.h>

#include "db_login.h"

/*-------------------------------------------------------------------------
 * create_new_database
 *
 * Create a new PostgreSQL database.
 *
 * Inputs:
 *    new_db : information about the new database.
 *    pgsu_data : the PosgreSQL super user information.  This user
 *                should already be connected.
 *
 * Returns the result of the database creation.  This value should be
 * examined and PQclear()'ed by the caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_new_database (DbLoginData *newdb_data,
                               DbLoginData  *pgsu_data);


/*-------------------------------------------------------------------------
 * create_sequences
 *
 * Create all of the sequences required.
 *
 * Inputs:
 *    data : information about the database to create the sequences in.
 *           The "data" should be connected to the database.
 *           
 * Returns the result of the last sequence the function attempted to
 * create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_sequences (DbLoginData *data, gint version);


/*-------------------------------------------------------------------------
 * create_tables
 *
 * Create all of the tables required.  
 *
 * Inputs:
 *    data : information about the database to create the tables in.
 *           The "data" should be connected to the database.
 *           
 * Returns the result of the last table the function attempted to
 * create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_tables (DbLoginData *data, gint version);


/*-------------------------------------------------------------------------
 * create_functions
 *
 * Create all of the external (library) functions.
 *
 * Inputs:
 *    data : information about the database to create the functions in.
 *           The "data" should be connected to the database.  This should
 *           be the superuser "data".
 *    version : The schema version.  Used to determine which functions
 *              are created and the version of the code to use.
 *
 * The version information is useful because this function is used 
 * both when creating a new schema and when updating an existing
 * schema (including repairs to old schemas).
 *           
 * Returns the result of the last function the function attempted to
 * create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_functions (DbLoginData *data, gint version);


/*-------------------------------------------------------------------------
 * create_languages
 *
 * Create all of the trusted languages.
 *
 * Inputs:
 *    data : information about the database to create the functions in.
 *           The "data" should be connected to the database.  This should
 *           be the superuser "data".
 *    version : The schema version.  Used to determine which languages
 *              are created and the version of the code to use.
 *
 * The version information is useful because this function is used 
 * both when creating a new schema and when updating an existing
 * schema (including repairs to old schemas).
 *           
 * Returns the result of the last language the function attempted to
 * create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_languages (DbLoginData *data, gint version);


/*-------------------------------------------------------------------------
 * create_stored_funcs
 *
 * Create all of the stored functions.
 *
 * Inputs:
 *    data : information about the database to create the functions in.
 *           The "data" should be connected to the database.  This should
 *           be the new user "data".
 *    version : The schema version.  Used to determine which functions
 *              are created and the version of the code to use.
 *
 * The version information is useful because this function is used 
 * both when creating a new schema and when updating an existing
 * schema (including repairs to old schemas).
 *           
 * Returns the result of the last stored functon the function attempted
 * to create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_stored_funcs (DbLoginData *data, gint version);


/*-------------------------------------------------------------------------
 * create_unique_indexes
 *
 * Create all of the unique indexes.
 *
 * Inputs:
 *    data : information about the database to create the functions in.
 *           The "data" should be connected to the database.  This should
 *           be the new user "data".
 *    version : The schema version.  Used to determine which indexes
 *              are created and the version of the code to use.
 *
 * The version information is useful because this function is used 
 * both when creating a new schema and when updating an existing
 * schema (including repairs to old schemas).
 *           
 * Returns the result of the last unique index the function attempted
 * to create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_unique_indexes (DbLoginData *data, gint version);


/*-------------------------------------------------------------------------
 * create_triggers
 *
 * Create all of the triggers.
 *
 * Inputs:
 *    data : information about the database to create the triggers in.
 *           The "data" should be connected to the database.  This should
 *           be the new user "data".
 *    version : The schema version.  Used to determine which triggers
 *              are created and the version of the code to use.
 *
 * The version information is useful because this function is used 
 * both when creating a new schema and when updating an existing
 * schema (including repairs to old schemas).
 *           
 * Returns the result of the last trigger the function attempted
 * to create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_triggers (DbLoginData *data, gint version);


/*-------------------------------------------------------------------------
 * create_rules
 *
 * Create all of the rules.
 *
 * Inputs:
 *    data : information about the database to create the rules in.
 *           The "data" should be connected to the database.  This should
 *           be the new user "data".
 *    version : The schema version.  Used to determine which rules
 *              are created and the version of the code to use.
 *
 * The version information is useful because this function is used 
 * both when creating a new schema and when updating an existing
 * schema (including repairs to old schemas).
 *           
 * Returns the result of the last rule the function attempted
 * to create.  This value should be examined and PQclear()'ed by the
 * caller.
 *-------------------------------------------------------------------------
 */
PGresult *create_rules (DbLoginData *data, gint version);
#endif
