/*-------------------------------------------------------------------------
 * Copyright (c) 2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_ddl
 *
 * Synopsis:
 *   Routines used to create the DDL strings u8sed by the other modules
 *   in this package.
 *    
 * $Id: db_ddl.c,v 1.4 2004/06/29 18:45:41 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libpq-fe.h>

#include "db_schema.h"

#define REVOKE "revoke all on \"%s\" from public"
#define GRANT  "grant all on \"%s\" to public"

static void
append_columns (GString *sql, GList *cols, gint version)
{
   GList      *iter;
   ColumnData *col_data;

   for (iter = g_list_first (cols);
         iter != NULL;
         iter = iter->next)
      {
      col_data = iter->data;
      if (col_data->minver <= version)
         {
         sql = g_string_append (sql, col_data->name->str);
         sql = g_string_append (sql, " ");
         sql = g_string_append (sql, pg_type_name (col_data->type));
         if (col_data->size > 0)
            {
            g_string_append_printf (sql, "(%u)", col_data->size);
            }
         if (!col_data->nullable)
            {
            sql = g_string_append (sql, " NOT NULL");
            }
         sql = g_string_append (sql, ", ");
         }
      }
}

static void
append_constraints (GString *sql, GList *cons, gint version)
{
   ConstraintData *cons_data;
   GList          *iter;

   /*
    * TODO: The constraints do not have a version.  We should add
    * one.
    */

   for (iter = g_list_first (cons);
         iter != NULL;
         iter = iter->next)
      {
      cons_data = iter->data;
      g_string_append_printf (sql, "CONSTRAINT %s %s, ",
                              cons_data->name->str,
                              cons_data->condition->str);
      }
}

static void
append_primary_key (GString *sql, GList *pk_cols)
{
   ColumnData *col;
   GList      *iter;

   if (pk_cols != NULL)
      {
      sql = g_string_append (sql, "PRIMARY KEY (");
      for (iter = g_list_first (pk_cols);
            iter != NULL;
            iter = iter->next)
         {
         col = iter->data;
         sql = g_string_append (sql, col->name->str);
         if (iter->next != NULL)
            {
            sql = g_string_append (sql, ", ");
            }
         }
      sql = g_string_append (sql, "), ");
      }
}

void
create_table_str (GString *buff, gint idx, gint version)
{
   TableData   *tab;

   g_assert (buff != NULL);

   tab = get_table (idx);
   g_assert (tab != NULL);
   g_assert (tab->columns != NULL);
   g_assert (tab->minver <= version);

   g_string_printf (buff, "CREATE TABLE %s (", tab->name->str);
   append_columns (buff, tab->columns, version);
   append_constraints (buff, tab->constraints, version);
   append_primary_key (buff, tab->pk_cols);

   /*
    * No matter what was appended last, we should have an extra
    * ", " at the end of the screen.  Change that comma to the
    * closing paren.
    */
   buff->str[buff->len - 2] = ')';
}


void
add_primary_key_str (GString *buff, gint idx, gint version)
{
   TableData    *tab;

   tab = get_table (idx);
   g_assert (tab != NULL);
   g_assert (tab->columns != NULL);
   g_assert (tab->pk_cols != NULL);
   g_assert (tab->minver <= version);

   g_string_printf (buff, "ALTER TABLE %s ADD ", tab->name->str);
   append_primary_key (buff, tab->pk_cols);

   /*
    * Do to the way we keep this consistent when buildin the
    * CREATE TABLE string, the string we just created will have an
    * extraneous ", " at the end.  Change the comma to a space.
    */
   buff->str[buff->len - 2] = ' ';
}


void
add_col_str (GString *buff, gint t_idx, gint c_idx)
{
   TableData   *tab;
   ColumnData  *col;
   GList       *iter;
   gint         i;

   g_assert (buff != NULL);

   tab = get_table (t_idx);
   g_assert (tab != NULL);
   g_assert (tab->columns != NULL);

   g_string_printf (buff, "ALTER TABLE %s ADD COLUMN ", tab->name->str);

   iter = g_list_first (tab->columns);
   for (i = 0; i < c_idx; i++)
      {
      g_assert (iter != NULL);
      iter = iter->next;
      }
   g_assert (iter != NULL);
   col = iter->data;

   g_string_append_printf (buff, "%s ", col->name->str);
   g_string_append_printf (buff, "%s ", pg_type_name (col->type));
   if (col->size > 0)
      {
      g_string_append_printf (buff, "(%u) ", col->size);
      }
}


void
col_not_null_str (GString *buff, gint t_idx, gint c_idx)
{
   TableData   *tab;
   ColumnData  *col;
   GList       *iter;
   gint         i;

   g_assert (buff != NULL);

   tab = get_table (t_idx);
   g_assert (tab != NULL);
   g_assert (tab->columns != NULL);

   g_string_printf (buff, "ALTER TABLE %s ALTER COLUMN ", tab->name->str);

   iter = g_list_first (tab->columns);
   for (i = 0; i < c_idx; i++)
      {
      g_assert (iter != NULL);
      iter = iter->next;
      }
   g_assert (iter != NULL);
   col = iter->data;

   g_string_append_printf (buff, "%s SET NOT NULL", col->name->str);
}


void
create_unique_index_str (GString *buff, UniqueIndexData *idx)
{
   g_assert (buff != NULL);
   g_assert (idx != NULL);

   g_string_printf (buff,
         "create unique index \"%s\" on \"%s\" using btree ( \"%s\" \"text_ops\")",
         idx->name->str, idx->table->str, idx->column->str);
}


PGresult *
fix_grants (PGconn *conn, const gchar *obj)
{
   ExecStatusType  status;
   PGresult       *res = NULL;
   GString        *sql;

   sql = g_string_new ("");

   g_string_printf (sql, REVOKE, obj);
   res = PQexec (conn, sql->str);
   status = PQresultStatus (res);
   if (status == PGRES_COMMAND_OK)
      {
      PQclear (res);
      g_string_printf (sql, GRANT, obj);
      res = PQexec (conn, sql->str);
      }

   g_string_free (sql, TRUE);
   return res;
}
