/*-------------------------------------------------------------------------
 * Copyright (c) 2003-2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_druid
 *
 * Synopsis:
 *    The "Database Druid" is used to gather the information required
 *    in order to create a new PRepS database.
 *    
 * $Id: db_druid.c,v 1.44 2004/11/20 03:14:06 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#ifdef HAVE_STATBUF_H
#  include <statbuf.h>
#endif
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>

#include "db_analyze.h"
#include "db_create.h"
#include "db_login.h"
#include "db_schema.h"
#include "db_update.h"
#include "gstr_utils.h"
#include "init_data.h"
#include "gtkutils.h"
#include "progdlg.h"

/*
 * Widgets
 */
#define DRUID                     "druid"
#define START_PAGE                "start_page"
#define PAGE_1                    "page1"
#define BRANCH_NEW_PAGE_1         "b1p1"
#define BRANCH_UPD_PAGE_1         "b2p1"
#define BRANCH_UPD_PAGE_2         "b2p2"
#define END_PAGE                  "end_page"
#define PGSU_ENTRY                "pgsu_entry"
#define PG_PASSWD_ENTRY           "pg_passwd_entry"
#define PORT_ENTRY                "port_entry"
#define HOST_ENTRY                "host_entry"
#define DBNAME_ENTRY              "dbname_entry"
#define DBDESCR_ENTRY             "dbdescr_entry"
#define DBOWNER_COMBO             "dbowner_combo"
#define DBOWNER_PASSWD_ENTRY      "dbowner_passwd_entry"
#define PLPGSQL_PATH_ENTRY        "plpgsql_path_entry"
#define LIBPREPS_PATH_ENTRY       "libpreps_path_entry"
#define CREATE_NEW_RBUTTON        "create_new_rbutton"
#define UPGRADE_RBUTTON           "upgrade_rbutton"
#define UPD_DBNAME_COMBO          "upd_dbname_combo"
#define UPD_DBOWNER_DISPLAY       "upd_dbowner_display"
#define UPD_DBOWNER_PASSWD_ENTRY  "upd_dbowner_passwd_entry"
#define UPD_INFO_DESCR_ENTRY      "upd_info_descr_entry"
#define UPD_PLPGSQL_PATH_ENTRY    "upd_plpgsql_path_entry"
#define UPD_PREPS_PATH_ENTRY      "upd_preps_path_entry"

/*
 * Object data
 */
#define PGSU_LOGIN_DATA           "pgsu_login_data"
#define PREPS_LOGIN_DATA          "preps_login_data"
#define NAME_LIST                 "name_list"
#define OWNER_LIST                "owner_list"
#define NELS                      "number_of_elements"

/*
 * Messages
 */
#define OPENING_MSG _("\
Welcome to the PRepS Database Tool.\n\
\n\
The PRepS Database Tool will gather the information required \
to either create a new PRepS database, or upgrade/fix an existing \
PRepS database.\n\
\n\
Before continuing, make sure you have the database access information (user \
id and password) for the PostgreSQL superuser (usually 'postgres'). \
Also, if you are creating a new database, make sure that the user you want \
to own the PRepS database has been properly set up in PostgreSQL.")

#define NEW_DATABASE_MSG _("A new PRepS database will be created as follows:")
#define UPD_DATABASE_MSG _("The PRepS database %s will be update as follows:")

#define PGSU_MISSING _("You must supply a PostgreSQL Superuser Id")
#define DBNAME_MISSING _("You must supply a database name")
#define DBDESCR_MISSING _("You must supply a database description")
#define PGSQL_MISSING _("You must supply a path to the PL/pgSQL library")
#define PREPS_MISSING _("You must supply a path to the PRepS library")
#define DB_EXISTS_ERR _("A database called '%s' already exists.")

#define USERS_SQL "select usename from pg_user where usecreatedb = TRUE"
#define DATABASE_SQL "\
select d.datname, u.usename \
  from pg_database d, pg_user u \
 where d.datdba = u.usesysid \
   and d.datname not in ('template0', 'template1')"
#define DB_EXISTS \
   "select 'x' from pg_database where upper (datname) = upper ('%s')"


static void
draw_update_end_page (GtkWidget *main_win)
{
   GnomeDruidPageEdge  *end_page;
   DbLoginData         *login_data;
   GString             *msg;
   GString             *tmp;

   end_page = GNOME_DRUID_PAGE_EDGE (lookup_widget (main_win, END_PAGE));
   g_assert (end_page != NULL);
   login_data = g_object_get_data (G_OBJECT (main_win), PREPS_LOGIN_DATA);
   g_assert (login_data != NULL);

   tmp = g_string_new ("");
   msg = g_string_new ("");
   g_string_append_printf (msg, UPD_DATABASE_MSG, login_data->db_name->str);

   g_string_append_printf (msg, _("\n\tDescription: %s"),
                           login_data->descr->str);
   msg = g_string_append (msg, _("\n\tCurrent Schema Version: "));
   g_string_append_printf (msg, "%d", get_db_version (login_data->conn));
   msg = g_string_append (msg, _("\n\tTarget Schema Version: "));
   g_string_append_printf (msg, "%d", TARGET_SCHEMA);

   msg = g_string_append (msg,
         _("\n\nPress the Apply button to update the database."));

   gnome_druid_page_edge_set_text (end_page, msg->str);
   gnome_druid_page_edge_set_title (end_page, _("Update PRepS Database"));
   g_string_free (msg, TRUE);
   g_string_free (tmp, TRUE);
}


static void
draw_create_end_page (GtkWidget *main_win)
{
   GnomeDruidPageEdge *end_page;
   DbLoginData        *login_data;
   GString            *msg;

   end_page = GNOME_DRUID_PAGE_EDGE (lookup_widget (main_win, END_PAGE));
   g_assert (end_page != NULL);
   login_data = g_object_get_data (G_OBJECT (main_win), PREPS_LOGIN_DATA);
   g_assert (login_data != NULL);

   msg = g_string_new (NEW_DATABASE_MSG);

   msg = g_string_append (msg, _("\n\tName: "));
   msg = g_string_append (msg, login_data->db_name->str);

   msg = g_string_append (msg, _("\n\tDescription: "));
   msg = g_string_append (msg, login_data->descr->str);

   msg = g_string_append (msg, _("\n\tOwner: "));
   msg = g_string_append (msg, login_data->username->str);

   if (login_data->host != NULL)
      {
      msg = g_string_append (msg, _("\n\tHost: "));
      msg = g_string_append (msg, login_data->host->str);
      }

   if (login_data->port != NULL)
      {
      msg = g_string_append (msg, _("\n\tPort: "));
      msg = g_string_append (msg, login_data->port->str);
      }

   msg = g_string_append (msg,
         _("\n\nPress the Apply button to create the database."));

   gnome_druid_page_edge_set_text (end_page, msg->str);
   gnome_druid_page_edge_set_title (end_page, _("Create PRepS Database"));
   g_string_free (msg, TRUE);
}

static void
destroy_gstr_list (GList *list)
{
   GList *iter;

   if (list != NULL)
      {
      for (iter = g_list_first (list);
            iter != NULL;
            iter = iter->next)
         {
         g_string_free ((GString *)iter->data, TRUE);
         }
      g_list_free (list);
      }
}


#ifdef USE_NEW_GTK
static void
fill_db_owner_combo (GtkWidget *parent,
                     PGconn    *conn)
{
   GtkComboBox    *db_owner_combo;
   GList          *name_list;
   PGresult       *res;
   gint            i;
   gint            n;
   gint            nels;
   gint            sel;
   GString        *curr_sel = NULL;
   GString        *item_str;

   db_owner_combo = GTK_COMBO_BOX (lookup_widget (parent, DBOWNER_COMBO));
   g_assert (db_owner_combo != NULL);

   /*
    * If there is a name list, this combo has already been populated.
    * Save the currently selected item so we can reselect it if it 
    * is in the newly repopulated combo.  Destroy the list and remove everything
    * from the old combo.
    */
   name_list = g_object_get_data (G_OBJECT (db_owner_combo), NAME_LIST);
   if (name_list != NULL)
      {
      sel = gtk_combo_box_get_active (db_owner_combo);
      if (sel != -1)
         {
         item_str = g_list_nth_data (name_list, sel);
         curr_sel = g_string_new (item_str->str);
         }
      destroy_gstr_list (name_list);
      name_list = NULL;
      nels = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (db_owner_combo),
                                                 NELS));
      for (i = 0; i < nels; i++)
         {
         gtk_combo_box_remove_text (db_owner_combo, nels - i - 1);
         }
      }

   res = PQexec (conn, USERS_SQL);
   if (chk_sql_ok (res, _("Fetch PostgreSQL Users")))
      {
      n = PQntuples (res);
      sel = 0;
      for (i = 0; i < n; i++)
         {
         /*
          * The item_str will be freed when the name_list is
          * destroyed.
          */
         item_str = g_string_new (PQgetvalue (res, i, 0));
         gtk_combo_box_append_text (db_owner_combo, item_str->str);
         name_list = g_list_append (name_list, item_str);
         if (curr_sel != NULL)
            {
            if (strcmp (curr_sel->str, item_str->str) == 0)
               {
               sel = i;
               }
            }
         }
      g_object_set_data (G_OBJECT (db_owner_combo), NAME_LIST, name_list); 
      g_object_set_data (G_OBJECT (db_owner_combo), NELS, GINT_TO_POINTER (n)); 
      gtk_combo_box_set_active (db_owner_combo, sel);
      }

   PQclear (res);
   if (curr_sel != NULL)
      {
      g_string_free (curr_sel, TRUE);
      }
}


static void
fill_db_name_combo (GtkWidget *parent,
                    PGconn    *conn)
{
   GtkComboBox    *db_name_combo;
   GtkEntry       *owner;
   GList          *name_list;
   GList          *owner_list;
   PGresult       *res;
   gint            i;
   gint            n;
   gint            nels;
   gint            sel;
   GString        *curr_sel = NULL;
   GString        *item_str;

   db_name_combo = GTK_COMBO_BOX (lookup_widget (parent, UPD_DBNAME_COMBO));
   g_assert (db_name_combo != NULL);
   owner = GTK_ENTRY (lookup_widget (parent, UPD_DBOWNER_DISPLAY));
   g_assert (owner != NULL);

   /*
    * If there is a name list, this combo has already been populated.
    * Save the currently selected item so we can reselect it if it 
    * is in the newly repopulated combo.  Destroy the list and remove
    * all existing items from the combo.
    */
   name_list = g_object_get_data (G_OBJECT (db_name_combo), NAME_LIST);
   if (name_list != NULL)
      {
      sel = gtk_combo_box_get_active (db_name_combo);
      if (sel != -1)
         {
         item_str = g_list_nth_data (name_list, sel);
         curr_sel = g_string_new (item_str->str);
         }
      destroy_gstr_list (name_list);
      name_list = NULL;
      nels = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (db_name_combo),
                                                 NELS));
      for (i = 0; i < nels; i++)
         {
         gtk_combo_box_remove_text (db_name_combo, nels - i - 1);
         }
      }
   owner_list = g_object_get_data (G_OBJECT (db_name_combo), OWNER_LIST);

   /*
    * The owner list should be in sync with the name list.  Destroy it
    * if it exists.
    */
   if (owner_list != NULL)
      {
      destroy_gstr_list (owner_list);
      owner_list = NULL;
      }

   res = PQexec (conn, DATABASE_SQL);
   if (chk_sql_ok (res, _("Fetch Database Names")))
      {
      n = PQntuples (res);
      sel = 0;
      for (i = 0; i < n; i++)
         {
         /*
          * The item_str will be freed when the name_list is
          * destroyed.
          */
         item_str = g_string_new (PQgetvalue (res, i, 0));
         gtk_combo_box_append_text (db_name_combo, item_str->str);
         name_list = g_list_append (name_list, item_str);
         if (curr_sel != NULL)
            {
            if (strcmp (curr_sel->str, item_str->str) == 0)
               {
               sel = i;
               }
            }
         item_str = g_string_new (PQgetvalue (res, i, 1));
         owner_list = g_list_append (owner_list, item_str);
         }
      g_object_set_data (G_OBJECT (db_name_combo), NAME_LIST, name_list); 
      g_object_set_data (G_OBJECT (db_name_combo), OWNER_LIST, owner_list); 
      g_object_set_data (G_OBJECT (db_name_combo), NELS, GINT_TO_POINTER (n));
      gtk_combo_box_set_active (db_name_combo, sel);
      item_str = g_list_nth_data (owner_list, sel);
      gtk_entry_set_text (owner, item_str->str);
      }

   PQclear (res);
   if (curr_sel != NULL)
      {
      g_string_free (curr_sel, TRUE);
      }
}
#else
static void
fill_db_owner_combo (GtkWidget *parent,
                     PGconn    *conn)
{
   GtkOptionMenu  *db_owner_menu;
   GtkWidget      *menu;
   GtkWidget      *item;
   GList          *name_list;
   PGresult       *res;
   gint            i;
   gint            n;
   gint            sel;
   GString        *curr_sel = NULL;
   GString        *item_str;

   db_owner_menu = GTK_OPTION_MENU (lookup_widget (parent, DBOWNER_COMBO));
   g_assert (db_owner_menu != NULL);

   /*
    * If there is a name list, this menu has already been populated.
    * Save the currently selected item so we can reselect it if it 
    * is in the newly created menu.  Destroy the list and the old menu.
    */
   name_list = g_object_get_data (G_OBJECT (db_owner_menu), NAME_LIST);
   if (name_list != NULL)
      {
      sel = gtk_option_menu_get_history (db_owner_menu);
      if (sel != -1)
         {
         item_str = g_list_nth_data (name_list, sel);
         curr_sel = g_string_new (item_str->str);
         }
      destroy_gstr_list (name_list);
      name_list = NULL;
      gtk_option_menu_remove_menu (db_owner_menu);
      }

   res = PQexec (conn, USERS_SQL);
   if (chk_sql_ok (res, _("Fetch PostgreSQL Users")))
      {
      n = PQntuples (res);
      menu = gtk_menu_new ();
      sel = 0;
      for (i = 0; i < n; i++)
         {
         /*
          * The item_str will be freed when the name_list is
          * destroyed.
          */
         item_str = g_string_new (PQgetvalue (res, i, 0));
         item = gtk_menu_item_new_with_label (item_str->str);
         gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
         gtk_widget_show (item);
         name_list = g_list_append (name_list, item_str);
         if (curr_sel != NULL)
            {
            if (strcmp (curr_sel->str, item_str->str) == 0)
               {
               sel = i;
               }
            }
         }
      g_object_set_data (G_OBJECT (db_owner_menu), NAME_LIST, name_list); 
      gtk_option_menu_set_menu (db_owner_menu, menu);
      gtk_option_menu_set_history (db_owner_menu, sel);
      }

   PQclear (res);
   if (curr_sel != NULL)
      {
      g_string_free (curr_sel, TRUE);
      }
}


static void
fill_db_name_combo (GtkWidget *parent,
                    PGconn    *conn)
{
   GtkOptionMenu  *db_name_menu;
   GtkEntry       *owner;
   GtkWidget      *menu;
   GtkWidget      *item;
   GList          *name_list;
   GList          *owner_list;
   PGresult       *res;
   gint            i;
   gint            n;
   gint            sel;
   GString        *curr_sel = NULL;
   GString        *item_str;

   db_name_menu = GTK_OPTION_MENU (lookup_widget (parent, UPD_DBNAME_COMBO));
   g_assert (db_name_menu != NULL);
   owner = GTK_ENTRY (lookup_widget (parent, UPD_DBOWNER_DISPLAY));
   g_assert (owner != NULL);

   /*
    * If there is a name list, this menu has already been populated.
    * Save the currently selected item so we can reselect it if it 
    * is in the newly created menu.  Destroy the list and the old menu.
    */
   name_list = g_object_get_data (G_OBJECT (db_name_menu), NAME_LIST);
   if (name_list != NULL)
      {
      sel = gtk_option_menu_get_history (db_name_menu);
      if (sel != -1)
         {
         item_str = g_list_nth_data (name_list, sel);
         curr_sel = g_string_new (item_str->str);
         }
      destroy_gstr_list (name_list);
      name_list = NULL;
      gtk_option_menu_remove_menu (db_name_menu);
      }
   owner_list = g_object_get_data (G_OBJECT (db_name_menu), OWNER_LIST);

   /*
    * The owner list should be in sync with the name list.  Destroy it
    * if it exists.
    */
   if (owner_list != NULL)
      {
      destroy_gstr_list (owner_list);
      owner_list = NULL;
      }

   res = PQexec (conn, DATABASE_SQL);
   if (chk_sql_ok (res, _("Fetch Database Names")))
      {
      n = PQntuples (res);
      menu = gtk_menu_new ();
      sel = 0;
      for (i = 0; i < n; i++)
         {
         /*
          * The item_str will be freed when the name_list is
          * destroyed.
          */
         item_str = g_string_new (PQgetvalue (res, i, 0));
         item = gtk_menu_item_new_with_label (item_str->str);
         gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
         gtk_widget_show (item);
         name_list = g_list_append (name_list, item_str);
         if (curr_sel != NULL)
            {
            if (strcmp (curr_sel->str, item_str->str) == 0)
               {
               sel = i;
               }
            }
         item_str = g_string_new (PQgetvalue (res, i, 1));
         owner_list = g_list_append (owner_list, item_str);
         }
      g_object_set_data (G_OBJECT (db_name_menu), NAME_LIST, name_list); 
      g_object_set_data (G_OBJECT (db_name_menu), OWNER_LIST, owner_list); 
      gtk_option_menu_set_menu (db_name_menu, menu);
      gtk_option_menu_set_history (db_name_menu, sel);
      item_str = g_list_nth_data (owner_list, sel);
      gtk_entry_set_text (owner, item_str->str);
      }

   PQclear (res);
   if (curr_sel != NULL)
      {
      g_string_free (curr_sel, TRUE);
      }
}
#endif


static void
on_druid_cancel (GnomeDruid  *druid, 
                 GtkWidget   *main_win)
{
   GtkToggleButton *upgbtn;
   gchar           *upgmsg = 
      _("Are you sure you want to quit without upgrading a database?");
   gchar           *cr8msg = 
      _("Are you sure you want to quit without creating a database?");
   gchar           *themsg;
   GtkWidget       *msg_dlg;

   upgbtn = GTK_TOGGLE_BUTTON (lookup_widget (main_win, UPGRADE_RBUTTON));
   g_assert (upgbtn != NULL);

   themsg = (gtk_toggle_button_get_active (upgbtn)) ? upgmsg : cr8msg;
   msg_dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                     GTK_MESSAGE_QUESTION,
                                     GTK_BUTTONS_YES_NO, themsg);
   if (gtk_dialog_run (GTK_DIALOG (msg_dlg)) == GTK_RESPONSE_YES)
      {
      gtk_widget_destroy (main_win);
      }
   gtk_widget_destroy (msg_dlg);
}


static gboolean
on_page_1_next (GnomeDruidPage *curr_page,
                GnomeDruid     *druid,
                GtkWidget      *main_win)
{
   GtkEntry        *user_entry;
   GtkEntry        *passwd_entry;
   GtkEntry        *host_entry;
   GtkEntry        *port_entry;
   GtkToggleButton *upgbtn;
   DbLoginData     *pgsu_data;
   GString         *pgsu_str;
   GString         *pgpass_str;
   GString         *host_str;
   GString         *port_str;
   GtkWidget       *dlg;
   gboolean         in_error = FALSE;

   user_entry = GTK_ENTRY (lookup_widget (main_win, PGSU_ENTRY));
   g_assert (user_entry != NULL);
   passwd_entry = GTK_ENTRY (lookup_widget (main_win, PG_PASSWD_ENTRY));
   g_assert (passwd_entry != NULL);
   host_entry = GTK_ENTRY (lookup_widget (main_win, HOST_ENTRY));
   g_assert (host_entry != NULL);
   port_entry = GTK_ENTRY (lookup_widget (main_win, PORT_ENTRY));
   g_assert (port_entry != NULL);
   upgbtn = GTK_TOGGLE_BUTTON (lookup_widget (main_win, UPGRADE_RBUTTON));
   g_assert (upgbtn != NULL);

   /*
    * Create the connection object if it has not already been created
    */
   pgsu_data = g_object_get_data (G_OBJECT (main_win), PGSU_LOGIN_DATA);
   if (pgsu_data == NULL)
      {
      pgsu_data = db_login_data_new ();
      g_object_set_data (G_OBJECT (main_win), PGSU_LOGIN_DATA, pgsu_data);
      }

   /*
    * Extract the data, and setup the connection object.
    */
   pgsu_str = g_string_new (gtk_entry_get_text (user_entry));
   g_string_strip (pgsu_str);
   pgpass_str = g_string_new (gtk_entry_get_text (passwd_entry));
   g_string_strip (pgpass_str);
   host_str = g_string_new (gtk_entry_get_text (host_entry));
   g_string_strip (host_str);
   port_str = g_string_new (gtk_entry_get_text (port_entry));
   g_string_strip (port_str);

   /*
    * Check that all required fields are filled in.
    */
   if (pgsu_str->len == 0)
      {
      /*
       * We need to put some sort of error out here...
       */
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    PGSU_MISSING);
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      in_error = TRUE;
      }

   /*
    * Unless there is a previous error, setup the connection and test it
    */ 
   if (!in_error)
      {
      set_db_name (pgsu_data, "template1");
      set_db_username (pgsu_data, pgsu_str->str);
      if (pgpass_str->len != 0)
         {
         set_db_password (pgsu_data, pgpass_str->str);
         }
      if (host_str->len != 0)
         {
         set_db_host (pgsu_data, host_str->str);
         }
      if (port_str->len != 0)
         {
         set_db_port (pgsu_data, port_str->str);
         }

      if (!db_connect (pgsu_data))
         {
         in_error = TRUE;
         dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                       GTK_MESSAGE_ERROR,
                                       GTK_BUTTONS_OK,
                                       PQerrorMessage (pgsu_data->conn));
         gtk_dialog_run (GTK_DIALOG (dlg));
         gtk_widget_destroy (dlg);
         db_disconnect (pgsu_data);
         }
      }

   g_string_free (pgsu_str, TRUE);
   g_string_free (pgpass_str, TRUE);
   g_string_free (host_str, TRUE);
   g_string_free (port_str, TRUE);

   if (!in_error)
      {
      GnomeDruidPage *page;

      if (gtk_toggle_button_get_active (upgbtn))
         {
         /*
          * TODO:  Create a function to count the number of databases on
          *        the server.  If there are no databases, we need to go 
          *        to the NEW page rather than the UPD page.  Give the
          *        users a message to let them know.
          */
         page = GNOME_DRUID_PAGE (lookup_widget (main_win, BRANCH_UPD_PAGE_1));
         g_assert (page != FALSE);
         }
      else
         {
         page = GNOME_DRUID_PAGE (lookup_widget (main_win, BRANCH_NEW_PAGE_1));
         g_assert (page != FALSE);
         }
      gnome_druid_set_page (druid, page);
      }

   return TRUE;
}


static void
on_branch_new_p1_prepare (GnomeDruidPage *page,
                          GnomeDruid     *druid,
                          GtkWidget      *main_win)
{
   DbLoginData *login_data;

   login_data = g_object_get_data (G_OBJECT (main_win), PGSU_LOGIN_DATA);
   g_assert (login_data != NULL);
   g_assert (login_data->conn != NULL);

   fill_db_owner_combo (main_win, login_data->conn);
}


static gboolean 
back_to_page_1 (GnomeDruidPage *curr_page,
                GnomeDruid     *druid,
                GtkWidget      *main_win)
{
   GnomeDruidPage  *page;

   page = GNOME_DRUID_PAGE (lookup_widget (main_win, PAGE_1));
   g_assert (page != NULL);

   gnome_druid_set_page (druid, page);

   return TRUE;
}

static gboolean
on_branch_new_p1_next (GnomeDruidPage *curr_page,
                       GnomeDruid     *druid,
                       GtkWidget      *main_win)
{
   DbLoginData   *login_data;
   DbLoginData   *newdb_login_data;
   GtkEntry      *name_entry;
   GtkEntry      *descr_entry;
   GtkEntry      *passwd_entry;
   GtkEntry      *pgsql_lib_entry;
   GtkEntry      *preps_lib_entry;
#ifdef USE_NEW_GTK
   GtkComboBox   *user_combo;
#else
   GtkOptionMenu *user_combo;
#endif
   GList         *user_list;
   GString       *name_str;
   GString       *descr_str;
   GString       *pgsql_str;
   GString       *preps_str;
   GString       *passwd_str;
   GtkWidget     *dlg;
   gboolean       in_error = FALSE;
   gint           sel;

   name_entry = GTK_ENTRY (lookup_widget (main_win, DBNAME_ENTRY));
   g_assert (name_entry != NULL);
   descr_entry = GTK_ENTRY (lookup_widget (main_win, DBDESCR_ENTRY));
   g_assert (descr_entry != NULL);
   passwd_entry = GTK_ENTRY (lookup_widget (main_win, DBOWNER_PASSWD_ENTRY));
   g_assert (passwd_entry != NULL);
   pgsql_lib_entry = GTK_ENTRY (lookup_widget (main_win,
                                               PLPGSQL_PATH_ENTRY));
   g_assert (pgsql_lib_entry != NULL);
   preps_lib_entry = GTK_ENTRY (lookup_widget (main_win,
                                               LIBPREPS_PATH_ENTRY));
   g_assert (preps_lib_entry != NULL);
   login_data = g_object_get_data (G_OBJECT (main_win), PGSU_LOGIN_DATA);
   g_assert (login_data != NULL);
   g_assert (login_data->conn != NULL);
#ifdef USE_NEW_GTK
   user_combo = GTK_COMBO_BOX (lookup_widget (main_win, DBOWNER_COMBO));
#else
   user_combo = GTK_OPTION_MENU (lookup_widget (main_win, DBOWNER_COMBO));
#endif
   g_assert (user_combo != NULL);

   name_str = g_string_new (gtk_entry_get_text (name_entry));
   g_string_strip (name_str);
   descr_str = g_string_new (gtk_entry_get_text (descr_entry));
   g_string_strip (descr_str);
   pgsql_str = g_string_new (gtk_entry_get_text (pgsql_lib_entry));
   g_string_strip (pgsql_str);
   preps_str = g_string_new (gtk_entry_get_text (preps_lib_entry));
   g_string_strip (preps_str);
   passwd_str = g_string_new (gtk_entry_get_text (passwd_entry));
   g_string_strip (passwd_str);

   /*
    * Verify that all required fields are filled in.
    */
   if (!in_error && name_str->len == 0)
      {
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    DBNAME_MISSING);
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      in_error = TRUE;
      }
   if (!in_error && descr_str->len == 0)
      {
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    DBDESCR_MISSING);
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      in_error = TRUE;
      }
   if (!in_error && pgsql_str->len == 0)
      {
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    PGSQL_MISSING);
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      in_error = TRUE;
      }
   if (!in_error && preps_str->len == 0)
      {
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    PREPS_MISSING);
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      in_error = TRUE;
      }

   /* 
    * Verify that a database with the given name does not already 
    * exist.
    */
   if (!in_error)
      {
      GString  *sql;
      PGresult *res;

      sql = g_string_new ("");
      g_string_printf (sql, DB_EXISTS, name_str->str);
      res = PQexec (login_data->conn, sql->str);
      if (chk_sql_ok (res, _("Determine if database exists")))
         {
         if (PQntuples (res) != 0)
            {
            GString *err_str;

            err_str = g_string_new ("");
            g_string_printf (err_str, DB_EXISTS_ERR, name_str->str);
            dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                          GTK_MESSAGE_ERROR,
                                          GTK_BUTTONS_OK,
                                          err_str->str);
            gtk_dialog_run (GTK_DIALOG (dlg));
            gtk_widget_destroy (dlg);
            g_string_free (err_str, TRUE);
            in_error = TRUE;
            }
         }
      else
         {
         in_error = TRUE;
         }

      PQclear (res);
      g_string_free (sql, TRUE);
      }

   /*
    * Update the login data for the new database.  Create a new object
    * if the new db login data has not been created yet.
    */
   newdb_login_data = g_object_get_data (G_OBJECT (main_win),
                                         PREPS_LOGIN_DATA);
   if (newdb_login_data == NULL)
      {
      newdb_login_data = db_login_data_new ();
      g_object_set_data (G_OBJECT (main_win),
                         PREPS_LOGIN_DATA, newdb_login_data);
      }
   set_db_descr (newdb_login_data, descr_str->str);
   set_db_password (newdb_login_data, passwd_str->str);
   if (login_data->host != NULL)
      {
      set_db_host (newdb_login_data, login_data->host->str);
      }
   if (login_data->port != NULL)
      {
      set_db_port (newdb_login_data, login_data->port->str);
      }

   user_list = g_object_get_data (G_OBJECT (user_combo), NAME_LIST);
   g_assert (user_list != NULL);
   user_list = g_list_first (user_list);
#ifdef USE_NEW_GTK
   sel = gtk_combo_box_get_active (user_combo);
#else
   sel = gtk_option_menu_get_history (user_combo);
#endif
   if (sel >= 0)
      {
      GString *user_str;

      user_str = g_list_nth_data (user_list, sel);
      set_db_username (newdb_login_data, user_str->str);
      }

   /* 
    * Temporarily set the db name to template1 to test the connection
    */
   set_db_name (newdb_login_data, "template1");
   if (!db_connect (newdb_login_data))
      {
      in_error = TRUE;
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    PQerrorMessage (newdb_login_data->conn));
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      }
   db_disconnect (newdb_login_data);
   set_db_name (newdb_login_data, name_str->str);

   g_string_free (name_str, TRUE);
   g_string_free (descr_str, TRUE);
   g_string_free (pgsql_str, TRUE);
   g_string_free (preps_str, TRUE);

   if (!in_error)
      {
      GnomeDruidPage *page;

      draw_create_end_page (main_win);
      page = GNOME_DRUID_PAGE (lookup_widget (main_win, END_PAGE));
      g_assert (page != NULL);
      gnome_druid_set_page (druid, page);
      }

   return TRUE;
}


static void
on_branch_upd_p1_prepare (GnomeDruidPage *page,
                          GnomeDruid     *druid,
                          GtkWidget      *main_win)
{
   DbLoginData *login_data;

   login_data = g_object_get_data (G_OBJECT (main_win), PGSU_LOGIN_DATA);
   g_assert (login_data != NULL);
   g_assert (login_data->conn != NULL);

   fill_db_name_combo (main_win, login_data->conn);
}


static gboolean
on_branch_upd_p1_next (GnomeDruidPage *page,
                       GnomeDruid     *druid,
                       GtkWidget      *main_win)
{
   DbLoginData    *login_data;
   DbLoginData    *pgsu_data;
#ifdef USE_NEW_GTK
   GtkComboBox    *db_combo;
#else
   GtkOptionMenu  *db_combo;
#endif
   GtkEntry       *passwd_entry;
   GtkWidget      *dlg;
   GList          *name_list;
   GList          *owner_list;
   gint            sel;
   GString        *owner;
   GString        *db;
   GString        *passwd;
   GString        *msg;
   gboolean        in_error = FALSE;

#ifdef USE_NEW_GTK
   db_combo = GTK_COMBO_BOX (lookup_widget (main_win, UPD_DBNAME_COMBO));
#else
   db_combo = GTK_OPTION_MENU (lookup_widget (main_win, UPD_DBNAME_COMBO));
#endif
   g_assert (db_combo != NULL);
   passwd_entry = GTK_ENTRY (lookup_widget (main_win,
                                            UPD_DBOWNER_PASSWD_ENTRY));
   g_assert (passwd_entry != NULL);
   pgsu_data = g_object_get_data (G_OBJECT (main_win), PGSU_LOGIN_DATA);
   g_assert (pgsu_data != NULL);
   g_assert (pgsu_data->conn != NULL);

   name_list = g_object_get_data (G_OBJECT (db_combo), NAME_LIST);
   g_assert (name_list != NULL);
   owner_list = g_object_get_data (G_OBJECT (db_combo), OWNER_LIST);
   g_assert (owner_list != NULL);

#ifdef USE_NEW_GTK
   sel = gtk_combo_box_get_active (db_combo);
#else
   sel = gtk_option_menu_get_history (db_combo);
#endif
   g_assert (sel >= 0);

   owner = g_list_nth_data (owner_list, sel);
   db = g_list_nth_data (name_list, sel);
   passwd = g_string_new (gtk_entry_get_text (passwd_entry));
   g_string_strip (passwd);

   login_data = g_object_get_data (G_OBJECT (main_win), PREPS_LOGIN_DATA);
   if (login_data == NULL)
      {
      login_data = db_login_data_new ();
      g_object_set_data (G_OBJECT (main_win), PREPS_LOGIN_DATA, login_data);
      }
   set_db_name (login_data, db->str);
   set_db_username (login_data, owner->str);
   set_db_password (login_data, passwd->str);
   if (pgsu_data->host != NULL)
      {
      set_db_host (login_data, pgsu_data->host->str);
      }
   if (pgsu_data->port != NULL)
      {
      set_db_port (login_data, pgsu_data->port->str);
      }

   if (!db_connect (login_data))
      {
      in_error = TRUE;
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    PQerrorMessage (login_data->conn));
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      db_disconnect (login_data);
      }

   /*
    * Analyze the database to see if it can be upgraded.
    */
   if (!in_error && !db_has_prereq_objects (login_data->conn))
      {
      msg = g_string_new (_("Required objects are missing from this database.  The database you are attempting to upgrade either is not a PRepS database, or has been damaged beyond repair."));

      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win),
                                    GTK_DIALOG_DESTROY_WITH_PARENT,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    msg->str);
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      g_string_free (msg, TRUE);
      in_error = TRUE;
      }

   g_string_free (passwd, TRUE);

   return in_error;
}


static void
on_branch_upd_p2_prepare (GnomeDruidPage *curr_page,
                          GnomeDruid     *druid,
                          GtkWidget      *main_win)
{
   GtkEntry     *entry;
   DbLoginData  *login_data;
   GString      *sql;
   PGresult     *res;
   gboolean      in_error;

   /*
    * Get the current library paths.  If there  are no libraries, use the
    * same default paths that we use when creating a new database.
    */
   login_data = g_object_get_data (G_OBJECT (main_win), PREPS_LOGIN_DATA);
   g_assert (login_data != NULL);
   g_assert (login_data->conn != NULL);

   sql = g_string_new ("");

   g_string_printf
      (sql, "SELECT distinct probin from pg_proc where proname in (%s)",
       "'plpgsql_call_handler'");

   res = PQexec (login_data->conn, sql->str); 
   in_error = !chk_sql_ok (res, _("Get PLpg/SQL Call Handler Path"));

   entry = GTK_ENTRY (lookup_widget (main_win, UPD_PLPGSQL_PATH_ENTRY));
   g_assert (entry != NULL);
   if (!in_error && PQntuples (res) > 0)
      {
      gtk_entry_set_text (entry, PQgetvalue (res, 0, 0));
      }
   else
      {
      gtk_entry_set_text (entry, PLPGSQL_LIB_PATH);
      }
   PQclear (res);

   g_string_printf
      (sql, "SELECT distinct probin from pg_proc where proname in (%s)",
       "'send_new_pr_msg', 'send_assign_msg', 'send_update_msg'");

   res = PQexec (login_data->conn, sql->str); 
   in_error = !chk_sql_ok (res, _("Get PRepS Library Path"));

   entry = GTK_ENTRY (lookup_widget (main_win, UPD_PREPS_PATH_ENTRY));
   g_assert (entry != NULL);
   if (!in_error && PQntuples (res) > 0)
      {
      gtk_entry_set_text (entry, PQgetvalue (res, 0, 0));
      }
   else
      {
      gtk_entry_set_text (entry, PREPS_LIB_PATH);
      }
   PQclear (res);

   g_string_free (sql, TRUE);
}


static gboolean
on_branch_upd_p2_next (GnomeDruidPage  *curr_page,
                       GnomeDruid      *druid,
                       GtkWidget       *main_win)
{
   GnomeDruidPage *page;
   DbLoginData    *login_data;
   GtkEntry       *descr_entry;
   GString        *descr_str;
   GtkWidget      *dlg;
   gboolean        in_error = FALSE;

   login_data = g_object_get_data (G_OBJECT (main_win), PREPS_LOGIN_DATA);
   g_assert (login_data != NULL);
   g_assert (login_data->conn != NULL);

   descr_entry = GTK_ENTRY (lookup_widget (main_win, UPD_INFO_DESCR_ENTRY));
   g_assert (descr_entry != NULL);

   descr_str = g_string_new (gtk_entry_get_text (descr_entry));
   g_string_strip (descr_str);

   if (!in_error && descr_str->len == 0)
      {
      dlg = gtk_message_dialog_new (GTK_WINDOW (main_win), 0,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    DBDESCR_MISSING);
      gtk_dialog_run (GTK_DIALOG (dlg));
      gtk_widget_destroy (dlg);
      in_error = TRUE;
      }
   set_db_descr (login_data, descr_str->str);

   if (!in_error)
      {
      draw_update_end_page (main_win);
      page = GNOME_DRUID_PAGE (lookup_widget (main_win, END_PAGE));
      g_assert (page != NULL);
      gnome_druid_set_page (druid, page);
      }

   return TRUE;
}


static gboolean
on_end_page_back (GnomeDruidPage *curr_page,
                  GnomeDruid     *druid,
                  GtkWidget      *main_win)
{
   GtkToggleButton  *btn;
   GnomeDruidPage   *page;

   btn = GTK_TOGGLE_BUTTON (lookup_widget (main_win, UPGRADE_RBUTTON));
   g_assert (btn != NULL);

   if (gtk_toggle_button_get_active (btn))
      {
      page = GNOME_DRUID_PAGE (lookup_widget (main_win, BRANCH_UPD_PAGE_2));
      g_assert (page != NULL);
      }
   else
      {
      page = GNOME_DRUID_PAGE (lookup_widget (main_win, BRANCH_NEW_PAGE_1));
      g_assert (page != NULL);
      }
   gnome_druid_set_page (druid, page);

   return TRUE;
}


/*
 * The database has changed, change the default owner.
 */
#ifdef USE_NEW_GTK
static void
on_db_changed (GtkComboBox *combo, GtkWidget *main_win)
{
   GList    *owner_list;
   GString  *owner;
   gint      sel;
   GtkEntry *owner_entry;

   owner_list = g_object_get_data (G_OBJECT (combo), OWNER_LIST);
   if (owner_list != NULL)
      {
      sel = gtk_combo_box_get_active (combo);
      if (sel >= 0)
         {
         owner = (GString *)g_list_nth_data (owner_list, sel);
         owner_entry = GTK_ENTRY (lookup_widget (main_win,
                                                 UPD_DBOWNER_DISPLAY));
         gtk_entry_set_text (owner_entry, owner->str);
         }
      }
}
#else
static void
on_db_changed (GtkOptionMenu *menu, GtkWidget *main_win)
{
   GList    *owner_list;
   GString  *owner;
   gint      sel;
   GtkEntry *owner_entry;

   owner_list = g_object_get_data (G_OBJECT (menu), OWNER_LIST);
   if (owner_list != NULL)
      {
      sel = gtk_option_menu_get_history (menu);
      if (sel >= 0)
         {
         owner = (GString *)g_list_nth_data (owner_list, sel);
         owner_entry = GTK_ENTRY (lookup_widget (main_win,
                                                 UPD_DBOWNER_DISPLAY));
         gtk_entry_set_text (owner_entry, owner->str);
         }
      }
}
#endif


static void
do_database_creation (GnomeDruid  *druid,
                      GtkWidget   *main_win)
{
   gboolean          in_error = FALSE;
   DbLoginData      *pgsu_data;
   DbLoginData      *newdb_data;
   PGresult         *res;
   GtkWidget        *dlg;
   GtkEntry         *pgsql_lib_entry;
   GtkEntry         *preps_lib_entry;
   GString          *pgsql_str;
   GString          *preps_str;
   GString          *template_db_name;
   GString          *msg;
   GtkWidget        *errdlg;

   pgsu_data = g_object_get_data (G_OBJECT (main_win), PGSU_LOGIN_DATA);
   g_assert (pgsu_data != NULL);
   newdb_data = g_object_get_data (G_OBJECT (main_win), PREPS_LOGIN_DATA);
   g_assert (newdb_data != NULL);
   pgsql_lib_entry = GTK_ENTRY (lookup_widget (main_win,
                                               PLPGSQL_PATH_ENTRY));
   g_assert (pgsql_lib_entry != NULL);
   preps_lib_entry = GTK_ENTRY (lookup_widget (main_win,
                                               LIBPREPS_PATH_ENTRY));
   g_assert (preps_lib_entry != NULL);

   dlg = create_prog_dlg(GTK_WINDOW (main_win), CREATE_DLG);
   gtk_widget_show (dlg);
   while (gtk_events_pending())
      {
      gtk_main_iteration();
      }

   /*
    * Reset the data the library paths in the data structures that
    * define the schema (just in case they have changed since 
    * the initialization).
    */
   pgsql_str = g_string_new (gtk_entry_get_text (pgsql_lib_entry));
   g_string_strip (pgsql_str);
   preps_str = g_string_new (gtk_entry_get_text (preps_lib_entry));
   g_string_strip (preps_str);
   set_creation_step_status (dlg, CREATE_DATABASE, STEP_RUNNING);
   while (gtk_events_pending())
      {
      gtk_main_iteration();
      }
   reset_library_paths(pgsql_str->str, preps_str->str);
   g_string_free (preps_str, TRUE);
   g_string_free (pgsql_str, TRUE);

   /*
    * Use the PostgreSQL superuser connection to create the database.
    * Do this while connected to the template database.
    */
   template_db_name = g_string_new (pgsu_data->db_name->str);
   res = create_new_database (newdb_data, pgsu_data);
   in_error = !chk_sql_ok (res, _("Create database"));
   PQclear (res);
   if (!in_error)
      {
      in_error = !db_connect (newdb_data);
      if (in_error)
         {
         set_creation_step_status (dlg, CREATE_DATABASE, STEP_FAILED);
         msg = g_string_new ("");
         g_string_printf (msg, _("Failed to connect to %s as owner."),
                          newdb_data->db_name->str);
         errdlg = gtk_message_dialog_new (GTK_WINDOW (dlg),
                                          GTK_DIALOG_DESTROY_WITH_PARENT,
                                          GTK_MESSAGE_ERROR,
                                          GTK_BUTTONS_OK,
                                          msg->str);               
         gtk_dialog_run (GTK_DIALOG (errdlg));
         gtk_widget_destroy (errdlg);
         g_string_free (msg, TRUE);
         }
      }


   /*
    * That was the only thing we needed to do while connected to the
    * template database.  Close that connection and log in to the
    * newly created database.
    */
   if (!in_error)
      {
      db_disconnect (pgsu_data);
      set_db_name (pgsu_data, newdb_data->db_name->str);
      in_error = !db_connect (pgsu_data);
      if (in_error)
         {
         set_creation_step_status (dlg, CREATE_DATABASE, STEP_FAILED);
         msg = g_string_new ("");
         g_string_printf
            (msg, _("Failed to connect to %s as PostgreSQL superuser."),
             newdb_data->db_name->str);
         errdlg = gtk_message_dialog_new (GTK_WINDOW (dlg),
                                          GTK_DIALOG_DESTROY_WITH_PARENT,
                                          GTK_MESSAGE_ERROR,
                                          GTK_BUTTONS_OK,
                                          msg->str);               
         gtk_dialog_run (GTK_DIALOG (errdlg));
         gtk_widget_destroy (errdlg);
         g_string_free (msg, TRUE);
         }
      }


   /*
    * Create the basic objects for the database.  This is done as
    * the database owner.  Connect to the new database.
    */
   if (!in_error)
      {
      set_creation_step_status (dlg, CREATE_DATABASE, STEP_SUCCESS);
      set_creation_step_status (dlg, CREATE_DB_STRUCTURE, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = create_sequences (newdb_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Create sequences"));
      PQclear (res);
      if (in_error)
         {
         set_creation_step_status (dlg, CREATE_DB_STRUCTURE, STEP_FAILED);
         }
      }
   if (!in_error)
      {
      res = create_tables (newdb_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Create tables"));
      if (in_error)
         {
         set_creation_step_status (dlg, CREATE_DB_STRUCTURE, STEP_FAILED);
         }
      PQclear (res);
      }
   if (!in_error)
      {
      res = create_unique_indexes (newdb_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Create unique indexes"));
      if (in_error)
         {
         set_creation_step_status (dlg, CREATE_DB_STRUCTURE, STEP_FAILED);
         }
      PQclear (res);
      }

   /*
    * Create the default admin user.  This MUST be done before the
    * triggers are in place, as some of the triggers check for the
    * current user being an admin before allowing inserts.  If we 
    * have no admin user, we can't do anything useful.
    */ 
   if (!in_error)
      {
      set_creation_step_status (dlg, CREATE_DB_STRUCTURE, STEP_SUCCESS);
      set_creation_step_status (dlg, CREATE_ADMIN_USER, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = insert_initial_admin_user (newdb_data);
      in_error = !chk_sql_ok (res, _("Create initial admin user"));
      if (in_error)
         {
         set_creation_step_status (dlg, CREATE_ADMIN_USER, STEP_FAILED);
         }
      PQclear (res);
      }
      

   /*
    * Load the PL/pgSQL language.  Load the functions that require
    * libPRepS.  This needs to be done by the PostgreSQL superuser.
    * Connect to the new database.
    */
   if (!in_error)
      {
      set_creation_step_status (dlg, CREATE_ADMIN_USER, STEP_SUCCESS);
      set_creation_step_status (dlg, INSTALL_LIBS, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = create_functions (pgsu_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Create functions"));
      PQclear (res);
      if (in_error)
         {
         set_creation_step_status (dlg, INSTALL_LIBS, STEP_FAILED);
         }
      }
   if (!in_error)
      {
      res = create_languages (pgsu_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Loading languages"));
      PQclear (res);
      if (in_error)
         {
         set_creation_step_status (dlg, INSTALL_LIBS, STEP_FAILED);
         }
      }

   /*
    * Load stored proceudres, create triggers.  This is done as the
    * database owner.  Connect to the new database.
    */
   if (!in_error)
      {
      set_creation_step_status (dlg, INSTALL_LIBS, STEP_SUCCESS);
      set_creation_step_status (dlg, INSTALL_PROCS, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = create_stored_funcs (newdb_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Create stored functions"));
      PQclear (res);
      if (in_error)
         {
         set_creation_step_status (dlg, INSTALL_PROCS, STEP_FAILED);
         }
      }
   if (!in_error)
      {
      res = create_triggers (newdb_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Create triggers"));
      PQclear (res);
      if (in_error)
         {
         set_creation_step_status (dlg, INSTALL_PROCS, STEP_FAILED);
         }
      }
   if (!in_error)
      {
      res = create_rules (newdb_data, TARGET_SCHEMA);
      in_error = !chk_sql_ok (res, _("Create rules"));
      PQclear (res);
      if (in_error)
         {
         set_creation_step_status (dlg, INSTALL_PROCS, STEP_FAILED);
         }
      }

   /*
    * Install some initial data to give the user a starting point.
    */
   if (!in_error)
      {
      set_creation_step_status (dlg, INSTALL_PROCS, STEP_SUCCESS);
      set_creation_step_status (dlg, INSTALL_INIT_DATA, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = insert_initial_data (newdb_data);
      in_error = !chk_sql_ok (res, _("Install Initial Data"));
      PQclear (res);
      if (in_error)
         {
         set_creation_step_status (dlg, INSTALL_INIT_DATA, STEP_FAILED);
         }
      }


   /*
    * If there were errors, drop the new schema.
    */
   if (!in_error)
      {
      set_creation_step_status (dlg, INSTALL_INIT_DATA, STEP_SUCCESS);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      }
   else
      {
      /*
       * Re-connect to the template database so we can drop the 
       * newly created database.
       */
      db_disconnect (pgsu_data);
      set_db_name (pgsu_data, template_db_name->str);
      in_error = !db_connect (pgsu_data);
      db_disconnect (newdb_data);
      drop_database (pgsu_data, newdb_data->db_name->str);
      }
   g_string_free (template_db_name, TRUE);
   activate_close_btn (dlg);
}


static void
do_database_update (GnomeDruid  *druid,
                    GtkWidget   *main_win)
{
   GtkWidget        *errdlg;
   GtkWidget        *dlg;
   DbLoginData      *pgsu_data;
   DbLoginData      *newdb_data;
   GString          *msg;
   gboolean          in_error = FALSE;
   PGresult         *res;

   pgsu_data = g_object_get_data (G_OBJECT (main_win), PGSU_LOGIN_DATA);
   g_assert (pgsu_data != NULL);
   newdb_data = g_object_get_data (G_OBJECT (main_win), PREPS_LOGIN_DATA);
   g_assert (newdb_data != NULL);

   dlg = create_prog_dlg(GTK_WINDOW (main_win), UPDATE_DLG);
   set_update_step_status (dlg, UPDATE_INIT, STEP_RUNNING);
   gtk_widget_show (dlg);
   while (gtk_events_pending())
      {
      gtk_main_iteration();
      }

   /*
    * The pgsu_data is likely still connected to the template database.
    * Close that connection, and log on the the database that is to
    * be updated.
    */
   db_disconnect (pgsu_data);
   set_db_name (pgsu_data, newdb_data->db_name->str);
   in_error = !db_connect (pgsu_data);
   if (in_error)
      {
      set_update_step_status (dlg, UPDATE_INIT, STEP_FAILED);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      msg = g_string_new ("");
      g_string_printf
         (msg, _("Failed to connect to %s as PostgreSQL superuser."),
          newdb_data->db_name->str);
      errdlg = gtk_message_dialog_new (GTK_WINDOW (dlg),
                                       GTK_DIALOG_DESTROY_WITH_PARENT,
                                       GTK_MESSAGE_ERROR,
                                       GTK_BUTTONS_OK,
                                       msg->str);               
      gtk_dialog_run (GTK_DIALOG (errdlg));
      gtk_widget_destroy (errdlg);
      g_string_free (msg, TRUE);
      }

   if (!in_error)
      {
      set_update_step_status (dlg, UPDATE_INIT, STEP_SUCCESS);
      set_update_step_status (dlg, DROP_OBJECTS, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = drop_indexes (newdb_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Drop Indexes"));
         if (in_error)
            {
            set_update_step_status (dlg, DROP_OBJECTS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      res = drop_primary_keys (newdb_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Drop Primary Keys"));
         if (in_error)
            {
            set_update_step_status (dlg, DROP_OBJECTS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      res = drop_rules (newdb_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Drop Rules"));
         if (in_error)
            {
            set_update_step_status (dlg, DROP_OBJECTS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      res = drop_triggers (newdb_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Drop Triggers"));
         if (in_error)
            {
            set_update_step_status (dlg, DROP_OBJECTS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      res = drop_stored_procedures (newdb_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Drop Stored Procedures"));
         if (in_error)
            {
            set_update_step_status (dlg, DROP_OBJECTS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      res = drop_languages (pgsu_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Drop Languages"));
         if (in_error)
            {
            set_update_step_status (dlg, DROP_OBJECTS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      res = drop_external_functions (pgsu_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Drop External Functions"));
         if (in_error)
            {
            set_update_step_status (dlg, DROP_OBJECTS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      set_update_step_status (dlg, DROP_OBJECTS, STEP_SUCCESS);
      set_update_step_status (dlg, UPDATE_SCHEMA, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = add_missing_tables (newdb_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Add New Tables"));
         if (in_error)
            {
            set_update_step_status (dlg, UPDATE_SCHEMA, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      res = add_missing_cols (newdb_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Add New Columns"));
         if (in_error)
            {
            set_update_step_status (dlg, UPDATE_SCHEMA, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            PQclear (res);
            }
         }
      res = add_missing_primary_keys (newdb_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Reset Primary Keys"));
         if (in_error)
            {
            set_update_step_status (dlg, UPDATE_SCHEMA, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            PQclear (res);
            }
         }
      res = add_missing_indexes (newdb_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Add missing indexes"));
         if (in_error)
            {
            set_update_step_status (dlg, UPDATE_SCHEMA, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            PQclear (res);
            }
         }
      }

   if (!in_error)
      {
      set_update_step_status (dlg, UPDATE_SCHEMA, STEP_SUCCESS);
      set_update_step_status (dlg, REINSTALL_LIBS, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = create_functions (pgsu_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Reinstall Libraries"));
         if (in_error)
            {
            set_update_step_status (dlg, REINSTALL_LIBS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      if (!in_error)
         {
         res = create_languages (pgsu_data, TARGET_SCHEMA);
         if (res != NULL)
            {
            in_error = !chk_sql_ok (res, _("Reinstall Languages"));
            if (in_error)
               {
               set_update_step_status (dlg, REINSTALL_LIBS, STEP_FAILED);
               while (gtk_events_pending())
                  {
                  gtk_main_iteration();
                  }
               }
            PQclear (res);
            }
         }
      }

   if (!in_error)
      {
      set_update_step_status (dlg, REINSTALL_LIBS, STEP_SUCCESS);
      set_update_step_status (dlg, REINSTALL_PROCS, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = create_stored_funcs (newdb_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Reinstall Stored Procedures"));
         if (in_error)
            {
            set_update_step_status (dlg, REINSTALL_PROCS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      set_update_step_status (dlg, REINSTALL_PROCS, STEP_SUCCESS);
      set_update_step_status (dlg, REINSTALL_TRIGS, STEP_RUNNING);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      res = create_triggers (newdb_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Reinstall Triggers"));
         if (in_error)
            {
            set_update_step_status (dlg, REINSTALL_TRIGS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   /*
    * We only have one rule, just lump its recreation in the same
    * step as the triggers...
    */
   if (!in_error)
      {
      res = create_rules (newdb_data, TARGET_SCHEMA);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Reinstall Rules"));
         if (in_error)
            {
            set_update_step_status (dlg, REINSTALL_TRIGS, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      set_update_step_status (dlg, REINSTALL_TRIGS, STEP_SUCCESS);
      set_update_step_status (dlg, UPDATE_VER_INFO, STEP_RUNNING);
      res = insert_version (newdb_data);
      if (res != NULL)
         {
         in_error = !chk_sql_ok (res, _("Insert Version Information"));
         if (in_error)
            {
            set_update_step_status (dlg, UPDATE_VER_INFO, STEP_FAILED);
            while (gtk_events_pending())
               {
               gtk_main_iteration();
               }
            }
         PQclear (res);
         }
      }

   if (!in_error)
      {
      set_update_step_status (dlg, UPDATE_VER_INFO, STEP_SUCCESS);
      while (gtk_events_pending())
         {
         gtk_main_iteration();
         }
      }
   activate_close_btn (dlg);
}


static void
on_apply (GnomeDruidPage *page,
          GnomeDruid     *druid,
          GtkWidget      *main_win)
{
   GtkToggleButton  *upgbtn;

   upgbtn = GTK_TOGGLE_BUTTON (lookup_widget (main_win, UPGRADE_RBUTTON));
   g_assert (upgbtn != NULL);

   if (gtk_toggle_button_get_active (upgbtn))
      {
      do_database_update (druid, main_win);
      }
   else
      {
      do_database_creation (druid, main_win);
      }
}


static void
on_window_show (GtkWidget *widget, gpointer user_data)
{
   GtkEntry  *entry;

   initialize_db_schema();

   entry = GTK_ENTRY (lookup_widget (widget, PLPGSQL_PATH_ENTRY));
   g_assert (entry != NULL);
   gtk_entry_set_text (entry, PLPGSQL_LIB_PATH);

   entry = GTK_ENTRY (lookup_widget (widget, LIBPREPS_PATH_ENTRY));
   g_assert (entry != NULL);
   gtk_entry_set_text (entry, PREPS_LIB_PATH);
}


static void
on_window_destroy (GObject *object, gpointer user_data)
{
   DbLoginData *login_data;
   GtkWidget   *widget;
   GList       *list;

   login_data = g_object_get_data (object, PGSU_LOGIN_DATA);
   if (login_data != NULL)
      {
      db_login_data_free (login_data);
      g_object_set_data (object, PGSU_LOGIN_DATA, NULL);
      }
   login_data = g_object_get_data (object, PREPS_LOGIN_DATA);
   if (login_data != NULL)
      {
      db_login_data_free (login_data);
      g_object_set_data (object, PREPS_LOGIN_DATA, NULL);
      }

   widget = lookup_widget (GTK_WIDGET (object), DBOWNER_COMBO);
   g_assert (widget != NULL);
   list = g_object_get_data (G_OBJECT (widget), NAME_LIST);
   destroy_gstr_list (list);

   widget = lookup_widget (GTK_WIDGET (object), UPD_DBNAME_COMBO);
   g_assert (widget != NULL);
   list = g_object_get_data (G_OBJECT (widget), NAME_LIST);
   destroy_gstr_list (list);
   list = g_object_get_data (G_OBJECT (widget), OWNER_LIST);
   destroy_gstr_list (list);

   cleanup_db_schema();

   gtk_main_quit();
}


GtkWidget *
create_db_druid (void)
{
   GtkWidget    *window;
   GtkWidget    *druid;
   GtkWidget    *startpage;
   GtkWidget    *page;
   GtkWidget    *page_vbox;
   GtkWidget    *endpage;
   GtkWidget    *vbox;
   GtkWidget    *hbox;
   GtkWidget    *label;
   GtkWidget    *entry;
   GtkWidget    *combo;
   GtkWidget    *sectvbox;
   GtkWidget    *spacehbox;
   GtkWidget    *rbutton;
   GSList       *rgroup;
   GtkSizeGroup *size_grp1;
   GtkSizeGroup *size_grp2;
   GtkSizeGroup *size_grp3;
   GtkSizeGroup *size_grp4;

   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title (GTK_WINDOW (window), _("PRepS Database Tool"));
   GLADE_HOOKUP_OBJECT_NO_REF (window, window, TOP_DOG);
   g_signal_connect (GTK_OBJECT (window), "show",
                     GTK_SIGNAL_FUNC (on_window_show), NULL);
   g_signal_connect (GTK_OBJECT (window), "destroy",
                     GTK_SIGNAL_FUNC (on_window_destroy), NULL);

   druid = gnome_druid_new ();
   gtk_widget_show (druid);
   gtk_container_add (GTK_CONTAINER (window), druid);
   gtk_container_set_border_width (GTK_CONTAINER (druid), 4);
   GLADE_HOOKUP_OBJECT (window, druid, DRUID);

   g_signal_connect (GTK_OBJECT (druid), "cancel",
                     GTK_SIGNAL_FUNC (on_druid_cancel), window);

   /*
    * The start page.
    */
   startpage = gnome_druid_page_edge_new (GNOME_EDGE_START);
   gtk_widget_show_all (startpage);
   gnome_druid_append_page (GNOME_DRUID (druid),
                            GNOME_DRUID_PAGE (startpage));
   gnome_druid_page_edge_set_title (GNOME_DRUID_PAGE_EDGE (startpage),
         PACKAGE_STRING);
   gnome_druid_page_edge_set_text (GNOME_DRUID_PAGE_EDGE (startpage),
                                   OPENING_MSG);
   GLADE_HOOKUP_OBJECT (window, startpage, START_PAGE);

   /*
    * Page 1 - Database Server Information
    */
   page = gnome_druid_page_standard_new ();
   gtk_widget_show_all (page);
   gnome_druid_append_page (GNOME_DRUID (druid), GNOME_DRUID_PAGE (page));
   gnome_druid_page_standard_set_title (GNOME_DRUID_PAGE_STANDARD (page),
         _("Database Server Information"));
   GLADE_HOOKUP_OBJECT (window, page, PAGE_1);
   g_signal_connect (GTK_OBJECT (page), "next",
                     GTK_SIGNAL_FUNC (on_page_1_next), window);

   size_grp1 = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
   
   page_vbox = GNOME_DRUID_PAGE_STANDARD (page)->vbox;
   gtk_container_set_border_width (GTK_CONTAINER (page_vbox),
                                   STD_DLG_BORDER);
   gtk_box_set_spacing (GTK_BOX (page_vbox), 2 * STD_BOX_SPACING);

   sectvbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (sectvbox);
   gtk_box_pack_start (GTK_BOX (page_vbox), sectvbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, sectvbox, "action_sectvbox");

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
         _("<span weight=\"bold\">Action</span>"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "action_label");
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (sectvbox), label, FALSE, FALSE, 0);

   spacehbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (spacehbox);
   gtk_box_pack_start (GTK_BOX (sectvbox), spacehbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, spacehbox, "action_spacehbox");

   label = gtk_label_new ("    ");
   gtk_box_pack_start (GTK_BOX (spacehbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   GLADE_HOOKUP_OBJECT (window, label, "action_spacelabel");

   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (vbox);
   gtk_box_pack_start (GTK_BOX (spacehbox), vbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, vbox, "action_vbox");

   rbutton = gtk_radio_button_new_with_label (NULL, _("Create new database"));
   rgroup = gtk_radio_button_get_group (GTK_RADIO_BUTTON (rbutton));
   gtk_widget_show (rbutton);
   gtk_box_pack_start (GTK_BOX (vbox), rbutton, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, rbutton, CREATE_NEW_RBUTTON);

   rbutton = gtk_radio_button_new_with_label (rgroup, 
         _("Update/Fix an existing database"));
   rgroup = gtk_radio_button_get_group (GTK_RADIO_BUTTON (rbutton));
   gtk_widget_show (rbutton);
   gtk_box_pack_start (GTK_BOX (vbox), rbutton, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, rbutton, UPGRADE_RBUTTON);

   sectvbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (sectvbox);
   gtk_box_pack_start (GTK_BOX (page_vbox), sectvbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, sectvbox, "conn_info_sectvbox");

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
         _("<span weight=\"bold\">Sever Information</span>"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "conn_info_label");
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (sectvbox), label, FALSE, FALSE, 0);

   spacehbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (spacehbox);
   gtk_box_pack_start (GTK_BOX (sectvbox), spacehbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, spacehbox, "conn_info_spacehbox");

   label = gtk_label_new ("    ");
   gtk_box_pack_start (GTK_BOX (spacehbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   GLADE_HOOKUP_OBJECT (window, label, "conn_info_spacelabel");

   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (vbox);
   gtk_box_pack_start (GTK_BOX (spacehbox), vbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, vbox, "conninfo_vbox");

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "pgsu_hbox");

   label = gtk_label_new (_("PostgreSQL Superuser Id:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp1, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "pgsu_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   gtk_entry_set_text (GTK_ENTRY (entry), "postgres");
   GLADE_HOOKUP_OBJECT (window, entry, PGSU_ENTRY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "pg_passwd_hbox");

   label = gtk_label_new (_("Password:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp1, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "pg_passwd_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
   GLADE_HOOKUP_OBJECT (window, entry, PG_PASSWD_ENTRY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "host_hbox");

   label = gtk_label_new (_("Host:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp1, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "host_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, HOST_ENTRY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "port_hbox");

   label = gtk_label_new (_("Port:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp1, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "port_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, PORT_ENTRY);

   /*
    * New Branch, Page 1  - PRepS Database Information
    */
   page = gnome_druid_page_standard_new ();
   gtk_widget_show_all (page);
   gnome_druid_append_page (GNOME_DRUID (druid), GNOME_DRUID_PAGE (page));
   gnome_druid_page_standard_set_title (GNOME_DRUID_PAGE_STANDARD (page),
         _("PRepS Database Information"));
   GLADE_HOOKUP_OBJECT (window, page, BRANCH_NEW_PAGE_1);
   g_signal_connect (GTK_OBJECT (page), "back",
                     GTK_SIGNAL_FUNC (back_to_page_1), window);
   g_signal_connect (GTK_OBJECT (page), "next",
                     GTK_SIGNAL_FUNC (on_branch_new_p1_next), window);
   g_signal_connect (GTK_OBJECT (page), "prepare",
                     GTK_SIGNAL_FUNC (on_branch_new_p1_prepare), window);

   size_grp2 = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

   page_vbox = GNOME_DRUID_PAGE_STANDARD (page)->vbox;
   gtk_container_set_border_width (GTK_CONTAINER (page_vbox),
                                   STD_DLG_BORDER);
   gtk_box_set_spacing (GTK_BOX (page_vbox), 2 * STD_BOX_SPACING);

   sectvbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (sectvbox);
   gtk_box_pack_start (GTK_BOX (page_vbox), sectvbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, sectvbox, "general_sectvbox");

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
         _("<span weight=\"bold\">General Information</span>"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "general_label");
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (sectvbox), label, FALSE, FALSE, 0);

   spacehbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (spacehbox);
   gtk_box_pack_start (GTK_BOX (sectvbox), spacehbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, spacehbox, "general_spacehbox");

   label = gtk_label_new ("    ");
   gtk_box_pack_start (GTK_BOX (spacehbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   GLADE_HOOKUP_OBJECT (window, label, "general_spacelabel");

   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (vbox);
   gtk_box_pack_start (GTK_BOX (spacehbox), vbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, vbox, "general_vbox");

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "dbname_hbox");

   label = gtk_label_new (_("Database Name:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp2, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "dbname_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, DBNAME_ENTRY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "dbdescr_hbox");

   label = gtk_label_new (_("Description:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp2, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "dbdescr_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, DBDESCR_ENTRY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "dbowner_hbox");

   label = gtk_label_new (_("Database Owner:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp2, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "dbowner_label");

#ifdef USE_NEW_GTK
   combo = gtk_combo_box_new_text ();
#else
   combo = gtk_option_menu_new ();
#endif
   gtk_widget_show (combo);
   gtk_box_pack_start (GTK_BOX (hbox), combo, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, combo, DBOWNER_COMBO);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "dbowner_passwd_hbox");

   label = gtk_label_new (_("Password:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp2, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "dbowner_passwd_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
   GLADE_HOOKUP_OBJECT (window, entry, DBOWNER_PASSWD_ENTRY);

   sectvbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (sectvbox);
   gtk_box_pack_start (GTK_BOX (page_vbox), sectvbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, sectvbox, "libpath_sectvbox");

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
         _("<span weight=\"bold\">Library Paths</span>"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "libpath_label");
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (sectvbox), label, FALSE, FALSE, 0);

   spacehbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (spacehbox);
   gtk_box_pack_start (GTK_BOX (sectvbox), spacehbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, spacehbox, "libpath_spacehbox");

   label = gtk_label_new ("    ");
   gtk_box_pack_start (GTK_BOX (spacehbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   GLADE_HOOKUP_OBJECT (window, label, "libpath_spacelabel");

   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (vbox);
   gtk_box_pack_start (GTK_BOX (spacehbox), vbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, vbox, "libpath_vbox");

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "plpgsql_path_hbox");

   label = gtk_label_new (_("PL/pgSQL Library:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp2, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "plpgsql_path_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, PLPGSQL_PATH_ENTRY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "libpreps_path_hbox");

   label = gtk_label_new (_("PRepS Library:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp2, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "libpreps_path_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, LIBPREPS_PATH_ENTRY);

   /*
    * Update Branch, Page 1  - PRepS Database Information
    */
   page = gnome_druid_page_standard_new ();
   gtk_widget_show_all (page);
   gnome_druid_append_page (GNOME_DRUID (druid), GNOME_DRUID_PAGE (page));
   gnome_druid_page_standard_set_title (GNOME_DRUID_PAGE_STANDARD (page),
         _("PRepS Database Information"));
   GLADE_HOOKUP_OBJECT (window, page, BRANCH_UPD_PAGE_1);
   g_signal_connect (GTK_OBJECT (page), "back",
                     GTK_SIGNAL_FUNC (back_to_page_1), window);
   g_signal_connect (GTK_OBJECT (page), "next",
                     GTK_SIGNAL_FUNC (on_branch_upd_p1_next), window);
   g_signal_connect (GTK_OBJECT (page), "prepare",
                     GTK_SIGNAL_FUNC (on_branch_upd_p1_prepare), window);

   size_grp3 = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

   page_vbox = GNOME_DRUID_PAGE_STANDARD (page)->vbox;
   gtk_container_set_border_width (GTK_CONTAINER (page_vbox),
                                   STD_DLG_BORDER);
   gtk_box_set_spacing (GTK_BOX (page_vbox), 2 * STD_BOX_SPACING);

   sectvbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (sectvbox);
   gtk_box_pack_start (GTK_BOX (page_vbox), sectvbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, sectvbox, "updgen_sectvbox");

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
         _("<span weight=\"bold\">General Information</span>"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "updgen_label");
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (sectvbox), label, FALSE, FALSE, 0);

   spacehbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (spacehbox);
   gtk_box_pack_start (GTK_BOX (sectvbox), spacehbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, spacehbox, "updgen_spacehbox");

   label = gtk_label_new ("    ");
   gtk_box_pack_start (GTK_BOX (spacehbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   GLADE_HOOKUP_OBJECT (window, label, "updgen_spacelabel");

   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (vbox);
   gtk_box_pack_start (GTK_BOX (spacehbox), vbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, vbox, "updgen_vbox");

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "upddbname_hbox");

   label = gtk_label_new (_("Database Name:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp3, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "upddbname_label");

#ifdef USE_NEW_GTK
   combo = gtk_combo_box_new_text ();
#else
   combo = gtk_option_menu_new ();
#endif
   gtk_widget_show (combo);
   gtk_box_pack_start (GTK_BOX (hbox), combo, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, combo, UPD_DBNAME_COMBO);
   g_signal_connect (combo, "changed",
                     G_CALLBACK (on_db_changed), window);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "upddbowner_hbox");

   label = gtk_label_new (_("Database Owner:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp3, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "upddbowner_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   gtk_widget_set_sensitive (entry, FALSE);
   GLADE_HOOKUP_OBJECT (window, entry, UPD_DBOWNER_DISPLAY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "upddbpasswd_hbox");

   label = gtk_label_new (_("Password:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp3, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "upddbowner_passwd_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
   GLADE_HOOKUP_OBJECT (window, entry, UPD_DBOWNER_PASSWD_ENTRY);

   /*
    * Update Branch, Page 2 - Library Information
    */
   page = gnome_druid_page_standard_new ();
   gtk_widget_show_all (page);
   gnome_druid_append_page (GNOME_DRUID (druid), GNOME_DRUID_PAGE (page));
   gnome_druid_page_standard_set_title (GNOME_DRUID_PAGE_STANDARD (page),
         _("Details"));
   GLADE_HOOKUP_OBJECT (window, page, BRANCH_UPD_PAGE_2);
   g_signal_connect (GTK_OBJECT (page), "next",
                     GTK_SIGNAL_FUNC (on_branch_upd_p2_next), window);
   g_signal_connect (GTK_OBJECT (page), "prepare",
                     GTK_SIGNAL_FUNC (on_branch_upd_p2_prepare), window);

   size_grp4 = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

   page_vbox = GNOME_DRUID_PAGE_STANDARD (page)->vbox;
   gtk_container_set_border_width (GTK_CONTAINER (page_vbox),
                                   STD_DLG_BORDER);
   gtk_box_set_spacing (GTK_BOX (page_vbox), 2 * STD_BOX_SPACING);

   sectvbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (sectvbox);
   gtk_box_pack_start (GTK_BOX (page_vbox), sectvbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, sectvbox, "updinfo_sectvbox");

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
         _("<span weight=\"bold\">Upgrade Information</span>"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "updinfo_label");
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (sectvbox), label, FALSE, FALSE, 0);

   spacehbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (spacehbox);
   gtk_box_pack_start (GTK_BOX (sectvbox), spacehbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, spacehbox, "updinfo_spacehbox");

   label = gtk_label_new ("    ");
   gtk_box_pack_start (GTK_BOX (spacehbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   GLADE_HOOKUP_OBJECT (window, label, "updinfo_spacelabel");

   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (vbox);
   gtk_box_pack_start (GTK_BOX (spacehbox), vbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, vbox, "updinfo_vbox");

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "upd_info_descr_hbox");

   label = gtk_label_new (_("Description:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp4, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "upd_info_descr_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, UPD_INFO_DESCR_ENTRY);

   sectvbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (sectvbox);
   gtk_box_pack_start (GTK_BOX (page_vbox), sectvbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, sectvbox, "updlibs_sectvbox");

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
         _("<span weight=\"bold\">Library Information</span>"));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "updlibs_label");
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (sectvbox), label, FALSE, FALSE, 0);

   spacehbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (spacehbox);
   gtk_box_pack_start (GTK_BOX (sectvbox), spacehbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, spacehbox, "updlibs_spacehbox");

   label = gtk_label_new ("    ");
   gtk_box_pack_start (GTK_BOX (spacehbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   GLADE_HOOKUP_OBJECT (window, label, "updlibs_spacelabel");

   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (vbox);
   gtk_box_pack_start (GTK_BOX (spacehbox), vbox, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, vbox, "updlibs_vbox");

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "upd_plpgsql_path_hbox");

   label = gtk_label_new (_("PL/pgSQL Library:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp4, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "upd_plpgsql_path_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, UPD_PLPGSQL_PATH_ENTRY);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   gtk_widget_show (hbox);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   GLADE_HOOKUP_OBJECT (window, hbox, "upd_preps_path_hbox");

   label = gtk_label_new (_("PRepS Library:"));
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_size_group_add_widget (size_grp4, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   GLADE_HOOKUP_OBJECT (window, label, "upd_preps_path_label");

   entry = gtk_entry_new ();
   gtk_widget_show (entry);
   gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
   GLADE_HOOKUP_OBJECT (window, entry, UPD_PREPS_PATH_ENTRY);

   /*
    * End page
    */
   endpage = gnome_druid_page_edge_new (GNOME_EDGE_FINISH);
   gtk_widget_show (endpage);
   gnome_druid_append_page (GNOME_DRUID (druid),
                            GNOME_DRUID_PAGE (endpage));
   g_signal_connect (GTK_OBJECT (endpage), "finish",
                     GTK_SIGNAL_FUNC (on_apply), window);
   g_signal_connect (GTK_OBJECT (endpage), "back",
                     GTK_SIGNAL_FUNC (on_end_page_back), window);
   GLADE_HOOKUP_OBJECT (window, endpage, END_PAGE);

   return window;
}
