/*-------------------------------------------------------------------------
 * Copyright (c) 2003 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_login
 *
 * Synopsis:
 *    Utility functions used for database access.
 *    
 * $Id: db_login.c,v 1.6 2003/11/26 03:11:54 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <syslog.h>
#include "db_login.h"

static gchar *
db_stat_string (ExecStatusType stat)
{
   switch (stat)
      {
      case PGRES_EMPTY_QUERY:
         return "Empty Query";
      case PGRES_COMMAND_OK:
         return "Command OK";
      case PGRES_TUPLES_OK:
         return "Tuples OK";
      case PGRES_COPY_OUT:
         return "Copy Out";
      case PGRES_COPY_IN:
         return "Copy In";
      case PGRES_BAD_RESPONSE:
         return "Bad Response";
      case PGRES_NONFATAL_ERROR:
         return "Nonfatal Error";
      case PGRES_FATAL_ERROR:
         return "Fatal Error";
      default:
         return "Unknown Error";
      }
}


void
set_db_username (DbLoginData *db_data, const char *username)
{
   if (username == NULL)
      {
      clear_db_username (db_data);
      }
   else
      {
      if (db_data->username != NULL)
         {
         g_string_assign (db_data->username, username);
         }
      else
         {
         db_data->username = g_string_new (username);
         }
      }
}

void
set_db_password (DbLoginData *db_data, const char *password)
{
   if (password == NULL)
      {
      clear_db_password (db_data);
      }
   else
      {
      if (db_data->password != NULL)
         {
         g_string_assign (db_data->password, password);
         }
      else
         {
         db_data->password = g_string_new (password);
         }
      }
}

void
set_db_name (DbLoginData *db_data, const char *name)
{
   if (name == NULL)
      {
      clear_db_name (db_data);
      }
   else
      {
      if (db_data->db_name != NULL)
         {
         g_string_assign (db_data->db_name, name);
         }
      else
         {
         db_data->db_name = g_string_new (name);
         }
      }
}

void
set_db_descr (DbLoginData *db_data, const char *descr)
{
   if (descr == NULL)
      {
      clear_db_descr (db_data);
      }
   else
      {
      if  (db_data->descr != NULL)
         {
         g_string_assign (db_data->descr, descr);
         }
      else
         {
         db_data->descr = g_string_new (descr);
         }
      }
}

void
set_db_host (DbLoginData *db_data, const char *host)
{
   if (host == NULL)
      {
      clear_db_host (db_data);
      }
   else
      {
      if (db_data->host != NULL)
         {
         g_string_assign (db_data->host, host);
         }
      else
         {
         db_data->host = g_string_new (host);
         }
      }
}

void
set_db_port (DbLoginData *db_data, const char *port)
{
   if (port == NULL)
      {
      clear_db_port (db_data);
      }
   else
      {
      if (db_data->port != NULL)
         {
         g_string_assign (db_data->port, port);
         }
      else
         {
         db_data->port = g_string_new (port);
         }
      }
}

void
clear_db_username (DbLoginData *db_data)
{
   if (db_data->username != NULL)
      {
      g_string_free (db_data->username, TRUE);
      db_data->username = NULL;
      }
}

void
clear_db_password (DbLoginData *db_data)
{
   if (db_data->password != NULL)
      {
      g_string_free (db_data->password, TRUE);
      db_data->password = NULL;
      }
}

void
clear_db_name (DbLoginData *db_data)
{
   if (db_data->db_name != NULL)
      {
      g_string_free (db_data->db_name, TRUE);
      db_data->db_name = NULL;
      }
}

void
clear_db_descr (DbLoginData *db_data)
{
   if (db_data->descr != NULL)
      {
      g_string_free (db_data->descr, TRUE);
      db_data->descr = NULL;
      }
}

void
clear_db_host (DbLoginData *db_data)
{
   if (db_data->host != NULL)
      {
      g_string_free (db_data->host, TRUE);
      db_data->host = NULL;
      }
}

void
clear_db_port (DbLoginData *db_data)
{
   if (db_data->port != NULL)
      {
      g_string_free (db_data->port, TRUE);
      db_data->port = NULL;
      }
}

DbLoginData *
db_login_data_new (void)
{
   DbLoginData *data;

   data = g_malloc (sizeof (DbLoginData));
   if (data != NULL)
      {
      data->username = NULL;
      data->password = NULL;
      data->db_name = NULL;
      data->descr = NULL;
      data->host = NULL;
      data->port = NULL;
      data->conn = NULL;
      }

   return data;
}

void
db_login_data_free (DbLoginData *db_data)
{
   db_disconnect (db_data);
   clear_db_username (db_data);
   clear_db_password (db_data);
   clear_db_name (db_data);
   clear_db_descr (db_data);
   clear_db_host (db_data);
   clear_db_port (db_data);

   g_free (db_data);
}

gboolean
db_connect (DbLoginData *db_data)
{
   GString *str;

   db_disconnect (db_data);
   str = g_string_new ("");
   if (db_data->db_name != NULL)
      {
      g_string_append_printf (str, "dbname='%s' ", db_data->db_name->str);
      }
   if (db_data->username != NULL)
      {
      g_string_append_printf (str, "user='%s' ", db_data->username->str);
      }
   if (db_data->password != NULL)
      {
      g_string_append_printf (str, "password='%s' ", db_data->password->str);
      }
   if (db_data->host != NULL)
      {
      g_string_append_printf (str, "host='%s' ", db_data->host->str);
      }
   if (db_data->port != NULL)
      {
      g_string_append_printf (str, "port='%s' ", db_data->port->str);
      }

   db_data->conn = PQconnectdb (str->str);
   g_string_free (str, TRUE);

   return (PQstatus (db_data->conn) != CONNECTION_BAD);
}

void
db_disconnect (DbLoginData *db_data)
{
   if (db_data->conn != NULL)
      {
      PQfinish (db_data->conn);
      }
   db_data->conn = NULL;
}


gboolean
chk_sql_ok (PGresult    *res,
            const gchar *descr_text)
{
   ExecStatusType  db_status;
   GString        *msg;
   GtkWidget      *msg_dlg;

   db_status = PQresultStatus (res);

   if (db_status != PGRES_TUPLES_OK && db_status != PGRES_COMMAND_OK)
      {
      msg = g_string_new (db_stat_string (db_status));
      msg = g_string_append_c (msg, '\n');
      msg = g_string_append (msg, descr_text);
      msg = g_string_append_c (msg, '\n');
      msg = g_string_append (msg, PQresultErrorMessage (res));
      msg_dlg = gtk_message_dialog_new (NULL, 0, 
                                        GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
                                        msg->str);
      gtk_dialog_run (GTK_DIALOG (msg_dlg));
      gtk_widget_destroy (msg_dlg);
      syslog (LOG_ERR, _("Error in SQL Execution: %s"), 
              PQresultErrorMessage (res));
      return FALSE;
      }

   return TRUE;
}


void
drop_database (DbLoginData *pgsu_data, const gchar *db_name)
{
   GString   *sql;
   PGresult  *res;

   sql = g_string_new ("DROP DATABASE ");
   sql = g_string_append (sql, db_name);

   res = PQexec (pgsu_data->conn, sql->str);
   PQclear (res);

   g_string_free (sql, TRUE);
}
