#ifndef __DB_SCHEMA_H
#define __DB_SCHEMA_H
/*-------------------------------------------------------------------------
 * Copyright (c) 2003-2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_schema
 *
 * Synopsis:
 *    Define the database schema.  If we start using new PostgreSQL
 *    featuures, this file will need to change.  If the schema is
 *    modified, the db_schema.c will need to change.
 *    
 * $Id: db_schema.h,v 1.13 2004/04/18 21:42:21 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */

/*
 * NOTES:
 *
 *   1.) Some structures have a prereq column.  If this column is TRUE,
 *       then an upgrade cannot be performed unless this object exists.
 *       In the case where the prereq flag is TRUE on a child object,
 *       but FALSE on the parent, the child is only required if the 
 *       parent object exists.
 *
 *   2.) Certain items contain min and max schema version flags.  These
 *       flags are used in "repair only" mode, to make sure that we only
 *       install items that are defined for the current version of the
 *       schema.
 */

#include <glib.h>

#define NOVER   -1

/*
 * This is NOT an inclusive list of pre-defined PostgreSQL types.  
 * Rather, it is limited to the types that we actually use.
 */
typedef enum {
   PG_BOOL,
   PG_INT4,
   PG_DATE,
   PG_TIMESTAMP,
   PG_CHRVAR,
   PG_NAME,
   PG_TEXT,
   PG_NAME_ARR,
   PG_INT4_ARR
} PgDataType;

gchar *pg_type_name (PgDataType type);

/*
 * Sequence data
 */
typedef struct SequenceData {
   GString *name;
   GString *ref_table_name;
   GString *ref_col_name;
   gulong   start;
   gulong   increment;
   gulong   minval;
   gulong   maxval;
   gulong   cache;
} SequenceData;

gint          count_sequences (void);
SequenceData *get_sequence (gint n);


/*
 * Table data
 */
typedef struct ConstraintData {
   GString    *name;
   GString    *condition;
} ConstraintData;

typedef struct ColumnData {
   GString    *name;
   PgDataType  type; 
   guint       size;
   gboolean    nullable;
   gint        minver;
   gboolean    prereq;        /* If true, this column must exist before we
                                 even attempt an upgrade */
   GString    *initsql;       /* When the column is added as part of an
                                 update, this SQL is used to initialize the
                                 column value */
} ColumnData;

typedef struct TableData {
   GString    *name;
   GList      *columns;
   GList      *pk_cols;
   GList      *constraints;
   gint        minver;
   gboolean    prereq;        /* If true, this table must exist before we
                                 even attempt an upgrade */
} TableData;

gint        count_tables (void);
TableData  *get_table (gint n);
gint        count_table_cols (const TableData *tab);
ColumnData *get_table_col (const TableData *tab, gint n);


/*
 * Language data
 */
typedef struct LanguageData {
   GString   *name;
   GString   *handler;
   GString   *compiler;
   gint       minver;
   gint       maxver;
} LanguageData;

gint          count_languages (void);
LanguageData *get_language (gint n);


/*
 * Function Data (external libs)
 */
typedef struct FunctionData {
   GString   *name;
   GString   *args;
   GString   *rtn;
   GString   *libpath;
   GString   *lang;
   gint       minver;
   gint       maxver;
} FunctionData;

gint          count_functions (void);
FunctionData *get_function (gint n);

/*
 * Stored Function Data (PLpg/SQL routines)
 */
typedef struct StoredFuncData {
   GString *name;
   GString *args;
   GString *rtn;
   GString *code;
   GString *lang;
   gint     minver;
   gint     maxver;
} StoredFuncData;

gint            count_stored_funcs (void);
StoredFuncData *get_stored_func (gint n);

/*
 * Unique Indexes
 */
typedef struct UniqueIndexData {
   GString *name;
   GString *table;
   GString *column;
   gint     minver;
   gint     maxver;
} UniqueIndexData;

gint             count_unique_indexes (void);
UniqueIndexData *get_unique_index (gint n);

/*
 * Trigger Information
 */
typedef struct TriggerData {
   GString  *name;
   gboolean  before;
   gboolean  ins;
   gboolean  upd;
   gboolean  del;
   GString  *table;
   gboolean  row;
   GString  *func;
   GString  *args;
   gint      minver;
   gint      maxver;
} TriggerData;

gint         count_triggers (void);
TriggerData *get_trigger (gint n);

/*
 * Rule Information
 */
typedef enum {
   SELECT_EVENT,
   UPDATE_EVENT,
   INSERT_EVENT,
   DELETE_EVENT
} PGEventType;

typedef struct RuleData {
   GString     *name;
   PGEventType  event;
   GString     *table;
   GString     *where;
   gboolean     instead;
   GString     *action;
   gint         minver;
   gint         maxver;
} RuleData;

gint      count_rules (void);
RuleData *get_rule (gint n);
/*
 * General functions.  Load the schema data into memory, and clean up
 * when done.
 */
void initialize_db_schema (void);
void cleanup_db_schema (void);
void reset_library_paths (const gchar *plpgsql_path,
                          const gchar *libpreps_path);

#endif
