/*-------------------------------------------------------------------------
 * Copyright (c) 2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * db_update
 *
 * Synopsis:
 *    Functions used to update a PRepS database.
 *    
 * $Id: db_update.c,v 1.8 2004/11/20 03:22:59 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdlib.h>

#include "db_analyze.h"
#include "db_ddl.h"
#include "db_update.h"
#include "db_schema.h"

#define FUNCTION_EXISTS "\
select proname \
  from pg_proc \
 where proname = '%s'"

#define SELECT_INDEXES "\
select c.relname \
  from pg_index i, pg_class c, pg_class t \
 where i.indisprimary = false \
   and c.oid = i.indexrelid \
   and t.oid = i.indrelid \
   and t.relname not like 'pg_%'"

#define SELECT_NARGS "\
select pronargs \
  from pg_proc \
 where proname = '%s'"

#define SELECT_PRIMARY_KEYS "\
select t.relname, c.relname \
  from pg_index i, pg_class c, pg_class t \
 where i.indisprimary = true \
   and c.oid = i.indexrelid \
   and t.oid = i.indrelid \
   and t.relname not like 'pg_%'"

#define SELECT_RULES "\
select rulename, tablename \
  from pg_rules \
 where tablename not like 'pg_%' "

#define SELECT_STORED_PROCEDURES "\
select p.proname, p.pronargs \
  from pg_proc p, pg_language l \
 where l.oid = p.prolang \
   and l.lanname = 'plpgsql' "

#define SELECT_SP_ARG "\
select typname \
  from pg_type \
 where oid = (select proargtypes[%d] \
                from pg_proc \
               where proname = '%s')"

#define SELECT_TRIGGERS "\
select t.tgname, c.relname \
  from pg_class c, pg_trigger t \
 where c.relname not like 'pg_%' \
   and c.oid = t.tgrelid"

#define DROP_INDEX        "drop index %s"
#define DROP_LANGUAGE     "drop language %s"
#define DROP_PRIMARY_KEY  "alter table %s drop constraint %s"
#define DROP_RULE         "drop rule %s on %s"
#define DROP_STORED_PROC  "drop function %s (%s)"
#define DROP_TRIGGER      "drop trigger %s on %s"


/*
 * To avoid repetative code, there are two generic drop functions.
 *
 * generic_drop() should be used for objects whose drop function only
 * require the the object name.  For example "DROP INDEX foo".
 *
 * generic_drop_2() should be used for objects whose drop function 
 * requires the object name plus another object name.  For example
 * "DROP TRIGGER foo ON bar"
 *
 * Arguments:
 *   conn - The database connection.
 *   sel  - The SQL to select all of the objects to be dropped.
 *          The SELECT clause should only contain the strings 
 *          used by the DROP template.  They should be in the 
 *          same order as they appear in the drop template (that
 *          is, for "DROP TRIGGER trigger_name ON table", the 
 *          SELECT should return the trigger_name and the table
 *          name in that order).
 *   drop - The DROP statement template that will be used.
 */
static PGresult *
generic_drop (PGconn *conn, const gchar *sel, const gchar *drop)
{
   GString  *sql;
   PGresult *res;
   PGresult *dropres = NULL;
   gint      i, n;

   res = PQexec (conn, sel);
   if (PQresultStatus (res) != PGRES_TUPLES_OK)
      {
      return res;
      }

   sql = g_string_new ("");
   n = PQntuples (res);
   for (i = 0; i < n; i++)
      {
      if (dropres != NULL)
         {
         PQclear (dropres);
         }
      g_string_printf (sql, drop, PQgetvalue (res, i, 0));
      dropres = PQexec (conn, sql->str);
      if (PQresultStatus (dropres) != PGRES_COMMAND_OK)
         {
         break;
         }
      }

   PQclear (res);
   g_string_free (sql, TRUE);

   return dropres;
}

static PGresult *
generic_drop_2 (PGconn *conn, const gchar *sel, const gchar *drop)
{
   GString  *sql;
   PGresult *res;
   PGresult *dropres = NULL;
   gint      i, n;

   res = PQexec (conn, sel);
   if (PQresultStatus (res) != PGRES_TUPLES_OK)
      {
      return res;
      }

   sql = g_string_new ("");
   n = PQntuples (res);
   for (i = 0; i < n; i++)
      {
      if (dropres != NULL)
         {
         PQclear (dropres);
         }
      g_string_printf (sql, drop, PQgetvalue (res, i, 0),
                       PQgetvalue (res, i, 1));
      dropres = PQexec (conn, sql->str);
      if (PQresultStatus (dropres) != PGRES_COMMAND_OK)
         {
         break;
         }
      }
   PQclear (res);
   g_string_free (sql, TRUE);

   return dropres;
}


static GString *
create_arg_list_string (PGconn *conn, const gchar *sp, gint nargs)
{
   GString  *sql;
   GString  *args;
   PGresult *res;
   gint      i;

   args = g_string_new ("");
   sql = g_string_new ("");

   for (i = 0; i < nargs; i++)
      {
      g_string_printf (sql, SELECT_SP_ARG, i, sp);
      res = PQexec (conn, sql->str);
      if (PQresultStatus (res) == PGRES_TUPLES_OK)
         {
         if (args->len > 0)
            {
            args = g_string_append (args, ", ");
            }
         args = g_string_append (args, PQgetvalue (res, 0, 0));
         }
      PQclear (res);
      }

   g_string_free (sql, TRUE);

   return args;
}


#if 0
PGresult *
drop_foreign_keys (DbLoginData *newdb_data)
{
}
#endif


PGresult *
drop_stored_procedures (DbLoginData *newdb_data)
{
   GString   *sql;
   GString   *args;
   PGconn    *conn = newdb_data->conn;
   PGresult  *res;
   PGresult  *dropres = NULL;
   gint       i, n;

   sql = g_string_new (SELECT_STORED_PROCEDURES);
   res = PQexec (conn, sql->str);
   n = PQntuples (res);
   for (i = 0; i < n; i++)
      {
      if (dropres != NULL)
         {
         PQclear (dropres);
         }
      args = create_arg_list_string (conn, PQgetvalue (res, i, 0),
                                     atoi (PQgetvalue (res, i, 1)));
      g_string_printf (sql, DROP_STORED_PROC, PQgetvalue (res, i, 0),
                       args->str); 
      g_string_free (args, TRUE);
      dropres = PQexec (conn, sql->str);
      if (PQresultStatus (dropres) != PGRES_COMMAND_OK)
         {
         break;
         }
      }
   PQclear (res);
   g_string_free (sql, TRUE);

   return dropres;
}


/*
 * For this one, there are only a limited number of functions that need
 * to be dropped, so just see if they exist, and then drop them.
 */
PGresult *
drop_external_functions (DbLoginData *pgsu_data)
{
   PGconn     *conn = pgsu_data->conn;
   GString    *sql;
   GString    *args;
   PGresult   *dropres = NULL;
   PGresult   *res = NULL;
   gint        i = 0;
   gint        nargs;
   gchar      *funcs[] = { "plpgsql_call_handler",
                           "send_new_pr_msg",
                           "send_assign_msg",
                           "send_update_msg",
                           NULL };

   sql = g_string_new ("");

   while (funcs[i] != NULL)
      {
      g_string_printf (sql, SELECT_NARGS, funcs[i]);
      res = PQexec (conn, sql->str);
      if (PQresultStatus (res) == PGRES_TUPLES_OK)
         {
         if (PQntuples (res) > 0)
            {
            nargs = atoi (PQgetvalue (res, 0, 0));
            args = create_arg_list_string (conn, funcs[i], nargs);
            g_string_printf (sql, DROP_STORED_PROC, funcs[i], args->str);
            g_string_free (args, TRUE);
            if (dropres != NULL)
               {
               PQclear (dropres);
               }
            dropres = PQexec (conn, sql->str);
            if (PQresultStatus (dropres) != PGRES_COMMAND_OK)
               {
               break;
               }
            }
         else
            {
            /*
             * If there was some sort of SQL error, return
             * the error so we can blow up and give something
             * to the user to report...
             */
            dropres = res;
            break;
            }
         }
      PQclear (res);
      }

   g_string_free (sql, TRUE);
   return dropres;
}


PGresult *
drop_languages (DbLoginData *pgsu_data)
{
   PGresult *res = NULL;
   PGconn   *conn = pgsu_data->conn;
   g_assert (conn != NULL);

   if (lang_exists (conn, "plpgsql"))
      {
      res = PQexec (conn, "drop language plpgsql");
      }

   return res;
}


PGresult *
drop_indexes (DbLoginData *newdb_data)
{
   PGconn   *conn = newdb_data->conn;
   g_assert (conn != NULL);
   return generic_drop (conn, SELECT_INDEXES, DROP_INDEX);
}


PGresult *
drop_primary_keys (DbLoginData *newdb_data)
{
   PGconn   *conn = newdb_data->conn;
   g_assert (conn != NULL);
   return generic_drop_2 (conn, SELECT_PRIMARY_KEYS, DROP_PRIMARY_KEY);
}


PGresult *
drop_rules (DbLoginData *newdb_data)
{
   PGconn   *conn = newdb_data->conn;
   g_assert (conn != NULL);
   return generic_drop_2 (conn, SELECT_RULES, DROP_RULE);
}


PGresult *
drop_triggers (DbLoginData *newdb_data)
{
   PGconn   *conn = newdb_data->conn;
   g_assert (conn != NULL);
   return generic_drop_2 (conn, SELECT_TRIGGERS, DROP_TRIGGER);
}


PGresult *
add_missing_indexes (DbLoginData *newdb_data, gint version)
{
   UniqueIndexData *index;
   gint             i;
   gint             n;
   GString         *sql;
   PGresult        *res = NULL;
   ExecStatusType   status;
   PGconn          *conn = newdb_data->conn;

   g_assert (conn != NULL);

   sql = g_string_new ("");

   /*
    * Look into changing this to just be indexes...  It seems
    * to me that any non-PK related index ought to be included.
    */
   n = count_unique_indexes();
   for (i = 0; i < n; i++)
      {
      index = get_unique_index (i);
      if (index->minver <= version &&
          (index->maxver >= version ||
           index->maxver == NOVER) &&
          !index_exists (conn, index->name->str))
          {
          create_unique_index_str (sql, index);
          if (res != NULL)
             {
             PQclear (res);
             }
          res = PQexec (conn, sql->str);
          status = PQresultStatus (res);
          if (status != PGRES_COMMAND_OK)
             {
             break;
             }
          }
      }
   g_string_free (sql, TRUE);

   return res;
}


PGresult *
add_missing_primary_keys (DbLoginData *newdb_data, gint version)
{
   TableData      *table;
   gint            i;
   gint            n;
   GString        *sql;
   PGresult       *res = NULL;
   ExecStatusType  status;
   PGconn         *conn = newdb_data->conn;

   g_assert (conn != NULL);

   sql = g_string_new ("");

   /*
    * The idea here is to return the result of the last PQexec().
    * If any of them fail, break at that point...
    *
    * Obviously, this function should only be called after
    * all tables have been created...
    *
    * Loop through all tables that exist.  Determine if the table 
    * should have a primary key.  If so, determine if the key exists
    * (we drop all keys as part of an update, so it should only exist
    * for brand new tables).  If it does not exist, create it.
    */
   n = count_tables();
   for (i = 0; i < n; i++)
      {
      table = get_table (i);
      if (table->minver <= version &&
            table_exists (conn, table->name->str) &&
            table->pk_cols != NULL &&
            !table_pk_exists (conn, table->name->str))
         {
         add_primary_key_str (sql, i, version);
         if (res != NULL)
            {
            PQclear (res);
            }
         res = PQexec (conn, sql->str);
         status = PQresultStatus (res);
         if (status != PGRES_COMMAND_OK)
            {
            break;
            }
         }
      }

   g_string_free (sql, TRUE);
   return res;
}


PGresult *
add_missing_tables (DbLoginData *newdb_data, gint version)
{
   TableData      *table;
   gint            i;
   gint            n;
   GString        *sql;
   PGresult       *res = NULL;
   ExecStatusType  status;
   PGconn         *conn = newdb_data->conn;

   g_assert (conn != NULL);

   sql = g_string_new ("");

   /*
    * The idea here is to return the result of the last PQexec().
    * If any of them fail, break at that point...
    */
   n = count_tables();
   for (i = 0; i < n; i++)
      {
      table = get_table (i);
      if (table->minver <= version &&
            !table_exists (conn, table->name->str))
         {
         create_table_str (sql, i, version);
         if (res != NULL)
            {
            PQclear (res);
            }
         res = PQexec (conn, sql->str);
         status = PQresultStatus (res);
         if (status != PGRES_COMMAND_OK)
            {
            break;
            }
         PQclear (res);
         res = fix_grants (conn, table->name->str);
         if (status != PGRES_COMMAND_OK)
            {
            break;
            }
         }
      }

   g_string_free (sql, TRUE);
   return res;
}


PGresult *
add_missing_cols (DbLoginData *newdb_data, gint version)
{
   TableData      *table;
   ColumnData     *column;
   gint            n;       /* number of tables */
   gint            i;       /* table index */
   gint            n_c;     /* number of columns */
   gint            i_c;     /* column index */
   GString        *sql;
   PGresult       *res = NULL;
   ExecStatusType  status;
   PGconn         *conn = newdb_data->conn;

   g_assert (conn != NULL);

   sql = g_string_new ("");

   /*
    * The idea here is to return the result of the last PQexec().
    * If any of them fail, break at that point...
    */
   n = count_tables();
   for (i = 0; i < n; i++)
      {
      table = get_table (i);
      if (table->minver <= version)
         {
         n_c = count_table_cols (table);
         for (i_c = 0; i_c < n_c; i_c++)
            {
            column = get_table_col (table, i_c);
            if (column->minver <= version &&
                  !col_exists (conn, table->name->str, column->name->str))
               {
               add_col_str (sql, i, i_c);
               if (res != NULL)
                  {
                  PQclear (res);
                  }
               res = PQexec (conn, sql->str);
               status = PQresultStatus (res);
               if (status != PGRES_COMMAND_OK)
                  {
                  break;
                  }
               if (column->initsql != NULL)
                  {
                  PQclear (res);
                  res = PQexec (conn, column->initsql->str);
                  status = PQresultStatus (res);
                  if (status != PGRES_COMMAND_OK &&
                      status != PGRES_TUPLES_OK)
                     {
                     break;
                     }
                  }
               if (!column->nullable && column->initsql != NULL)
                  {
                  col_not_null_str (sql, i, i_c);
                  PQclear (res);
                  res = PQexec (conn, sql->str);
                  status = PQresultStatus (res);
                  if (status != PGRES_COMMAND_OK)
                     {
                     break;
                     }
                  }
               }
            }
         }
      }

   g_string_free (sql, TRUE);
   return res;
}
