/*-------------------------------------------------------------------------
 * Copyright (c) 2003-2004 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * defdata
 *
 * Synopsis:
 *    Define the default data (that is, the initial data that is required
 *    to have a minimally functional PRepS database).
 *    
 * $Id: init_data.c,v 1.3 2004/08/05 03:33:58 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include "gstr_utils.h"
#include "init_data.h"

#define CREATE_SEVR_NUM_FUNC "\
CREATE FUNCTION get_severity_number (text) RETURNS int4 AS' \
    DECLARE \
        sevr_name ALIAS FOR $1; \
        sevr_num  int4; \
    BEGIN \
        SELECT severity_num \
          INTO sevr_num \
          FROM severity \
         WHERE name = sevr_name; \
        RETURN sevr_num; \
    END; \
' LANGUAGE 'plpgsql'"
#define DROP_SEVR_NUM_FUNC "DROP FUNCTION get_severity_number (text)"

#define STATUS_TEMPLATE "\
INSERT INTO status (status_num, name, order_num, active, is_closed_status, \
                    descr) \
VALUES (nextval ('status_seq'), '%s', %d, %s, %s, '%s')"

#define SEVR_TEMPLATE "\
INSERT INTO severity (severity_num, name, order_num, active, descr) \
     VALUES (nextval ('severity_seq'), '%s', %d, %s, '%s')"

#define PROB_TYPE_TEMPLATE "\
INSERT INTO problem_type (type_num, name, severity_num, active, descr) \
     VALUES (nextval ('problem_type_seq'), '%s', \
             get_severity_number ('%s'), %s, '%s')"

#define PR_QUERY_TEMPLATE "\
INSERT INTO pr_query (query_num, name, descr, is_system_query, \
                      is_raw_sql, order_by) \
     VALUES (nextval ('pr_query_seq'), '%s', '%s', TRUE, FALSE, '%s')"

#define PR_QUERY_WITH_STAT_TEMPLATE "\
INSERT INTO pr_query (query_num, name, descr, is_system_query, \
                      is_raw_sql, status_nums, order_by) \
     VALUES (nextval ('pr_query_seq'), '%s', '%s', TRUE, FALSE, '%s', '%s')"

#define RAW_QUERY_TEMPLATE "\
INSERT INTO pr_query (query_num, name, descr, is_system_query, \
                      is_raw_sql, raw_sql) \
     VALUES (nextval ('pr_query_seq'), '%s', '%s', TRUE, TRUE, '%s')"

static PGresult *
create_utility_functions (DbLoginData *db_data)
{
   return PQexec (db_data->conn, CREATE_SEVR_NUM_FUNC);
}

static PGresult *
drop_utility_functions (DbLoginData *db_data)
{
   return PQexec (db_data->conn, DROP_SEVR_NUM_FUNC);
}

static PGresult *
insert_statuses (DbLoginData *db_data)
{
   ExecStatusType  status;
   GString        *sql;
   PGresult       *res;
   gboolean        in_error;

   sql = g_string_new ("");

   g_string_printf (sql, STATUS_TEMPLATE, "new", 5, "TRUE", "FALSE",
                    "Report is new and unexamined.");
   res = PQexec (db_data->conn, sql->str);
   status = PQresultStatus (res);
   in_error = (status != PGRES_COMMAND_OK);

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, STATUS_TEMPLATE, "open", 10, "TRUE", "FALSE", 
         "Report has been examined and verified to be a problem.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, STATUS_TEMPLATE, "on hold", 15, "TRUE", "FALSE",
         "Waiting for feedback of some kind.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, STATUS_TEMPLATE, "fixed", 20, "TRUE", "FALSE",
         "Problem has been fixed, and is ready to be tested.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, STATUS_TEMPLATE, "not a problem", 25, "TRUE",
         "TRUE", "The report was found to not be a problem.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, STATUS_TEMPLATE, "closed", 30, "TRUE", "TRUE",
         "Problem has been fixed, and the solution verified.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   g_string_free (sql, TRUE);
   return res;
}

static PGresult *
insert_severities (DbLoginData *db_data)
{
   ExecStatusType  status;
   GString        *sql;
   PGresult       *res;
   gboolean        in_error;

   sql = g_string_new ("");

   g_string_printf (sql, SEVR_TEMPLATE, "non-essential", 25, "TRUE",
      "Use for suggestions, or problems with simple work arounds.");
   res = PQexec (db_data->conn, sql->str);
   status = PQresultStatus (res);
   in_error = (status != PGRES_COMMAND_OK);

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, SEVR_TEMPLATE, "low", 20, "TRUE", 
         "The problem has a simple work around, or very litte effect.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, SEVR_TEMPLATE, "medium", 15, "TRUE", 
         "There is a non-trivial work around for the problem.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, SEVR_TEMPLATE, "high", 10, "TRUE", 
         "There is no reasonable work around for the problem.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, SEVR_TEMPLATE, "essential", 5, "TRUE", 
         "The problem is a real show stopper.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   g_string_free (sql, TRUE);
   return res;
}

static GString *
build_status_key_list (PGconn *conn, gchar *statuses)
{
   ExecStatusType  status;
   GString        *sql;
   GString        *keys;
   PGresult       *res;
   gint            i, n;

   sql = g_string_new ("select status_num from status where name in ");
   g_string_append_printf (sql, "(%s)", statuses);
   res = PQexec (conn, sql->str);
   status = PQresultStatus (res);
   g_assert (status == PGRES_TUPLES_OK);

   n = PQntuples (res);
   g_assert (n > 0);
   keys = g_string_new ("{");
   keys = g_string_append (keys, PQgetvalue (res, 0, 0));
   for (i = 1; i < n; i++)
      {
      keys = g_string_append (keys, ",");
      keys = g_string_append (keys, PQgetvalue (res, i, 0));
      }

   keys = g_string_append (keys, "}");

   g_string_free (sql, TRUE);
   PQclear (res);
   
   return keys;
}



static PGresult *
insert_problem_types (DbLoginData *db_data)
{
   ExecStatusType  status;
   GString        *sql;
   PGresult       *res;
   gboolean        in_error;

   sql = g_string_new ("");

   g_string_printf (sql, PROB_TYPE_TEMPLATE, "sw-bug", "high", "TRUE",
      "There is a bug with the software.");
   res = PQexec (db_data->conn, sql->str);
   status = PQresultStatus (res);
   in_error = (status != PGRES_COMMAND_OK);

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, PROB_TYPE_TEMPLATE, "doc-bug", "medium", "TRUE", 
         "There is a problem with the documentation.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, PROB_TYPE_TEMPLATE, "suggestion", "non-essential",
         "TRUE", "Suggestions and comments.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, PROB_TYPE_TEMPLATE, "feature-req", "medium",
         "TRUE", "Request for a feature in the software.");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   g_string_free (sql, TRUE);
   return res;
}

static PGresult *
insert_queries (DbLoginData *db_data)
{
   ExecStatusType  status;
   GString        *sql;
   GString        *stats;
   PGresult       *res;
   gboolean        in_error;

   sql = g_string_new ("");

   g_string_printf (sql, PR_QUERY_TEMPLATE, "<ALL>",
      "Show all problem reports for the selected project",
      "prj.name, pr.problem_num");
   res = PQexec (db_data->conn, sql->str);
   status = PQresultStatus (res);
   in_error = (status != PGRES_COMMAND_OK);

   if (!in_error)
      {
      PQclear (res);
      stats = build_status_key_list (db_data->conn, "'new'");
      g_string_printf (sql, PR_QUERY_WITH_STAT_TEMPLATE, "New Problem Reports",
         "Show problem reports with a status of \"new\"",
         stats->str, "prj.name, pr.problem_num");
      g_string_free (stats, TRUE);
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      stats = build_status_key_list (db_data->conn, "'closed'");
      g_string_printf (sql, PR_QUERY_WITH_STAT_TEMPLATE,
         "Closed Problem Reports",
         "Show problem reports with a status of \"closed\"",
         stats->str, "prj.name, pr.problem_num");
      g_string_free (stats, TRUE);
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      stats = build_status_key_list (db_data->conn, "'fixed', 'not a problem'");
      g_string_printf (sql, PR_QUERY_WITH_STAT_TEMPLATE,
         "Ready for Verification",
         "Show problem reports that are ready for verification before closure.",
         stats->str, "prj.name, pr.problem_num");
      g_string_free (stats, TRUE);
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      g_string_printf (sql, RAW_QUERY_TEMPLATE, "Unassigned Problems",
            "Show problem reports that are not assigned to anyone else.",
            "and pr.responsible_id is NULL ORDER BY prj.name, "
            "stat.order_num, sevr.order_num, pr.problem_num");
      res = PQexec (db_data->conn, sql->str);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   g_string_free (sql, TRUE);
   return res;
}

PGresult *
insert_version (DbLoginData *db_data)
{
   GString   *sql;
   GString   *descr;
   PGresult  *res;

   descr = g_string_new (db_data->descr->str);
   g_string_prepare_db_instr (descr);

   sql = g_string_new
      ("insert into version (schema_ver, name, descr, install_date) ");
   g_string_append_printf (sql, "values (%d, '%s', %s, now())",
                           TARGET_SCHEMA, db_data->db_name->str, descr->str);
   g_string_free (descr, TRUE);

   res = PQexec (db_data->conn, sql->str);
   g_string_free (sql, TRUE);

   return res;
}


PGresult *
insert_initial_admin_user (DbLoginData *db_data)
{
   GString  *sql;
   PGresult *res;

   sql = g_string_new ("INSERT INTO person (login_id, last_name, ");
   sql = g_string_append (sql, "first_name, active, access_flag, ");
   sql = g_string_append (sql, "creator, creation_date) ");
   sql = g_string_append (sql, "VALUES (getpgusername(), 'Admin', ");
   sql = g_string_append (sql, "'Default', TRUE, 1, " );
   sql = g_string_append (sql, "getpgusername(), now())" );

   res = PQexec (db_data->conn, sql->str);

   g_string_free (sql, TRUE);

   return res;
}


PGresult *
insert_initial_data (DbLoginData *db_data)
{
   ExecStatusType  status;
   PGresult       *res;
   gboolean        in_error;

   res = create_utility_functions (db_data);
   status = PQresultStatus (res);
   in_error = (status != PGRES_COMMAND_OK);

   if (!in_error)
      {
      PQclear (res);
      res = insert_statuses (db_data);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      res = insert_severities (db_data);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      res = insert_problem_types (db_data);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      res = insert_queries (db_data);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      res = insert_version (db_data);
      status = PQresultStatus (res);
      in_error = (status != PGRES_COMMAND_OK);
      }

   if (!in_error)
      {
      PQclear (res);
      res = drop_utility_functions (db_data);
      }
   return res;
}
